from .context import estraces  # noqa
import numpy as np
import os
from estraces.formats.buffered_ets_writer import BufferedETSWriter, _Dataset, _SAMPLES_KEY
from estraces import read_ths_from_ets_file
import pytest

_output_ets_filename = 'BUFFERED_ETS-test.ets'


@pytest.fixture
def ets_reader():
    return read_ths_from_ets_file('tests/samples/test.ets')


@pytest.fixture
def output_ets_filename():
    try:
        os.remove(_output_ets_filename)
    except FileNotFoundError:
        pass
    yield _output_ets_filename
    try:
        os.remove(_output_ets_filename)
    except FileNotFoundError:
        pass


#############
#  Dataset  #
#############

def test_dataset_instantiation():
    _Dataset(42)


def test_dataset_raises_if_write_greater_than_1d_data():
    d = _Dataset(42)
    with pytest.raises(ValueError, match='Data must'):
        d.write(np.zeros((2, 3)))


def test_dataset_raises_if_write_when_full():
    d = _Dataset(42)
    data = np.zeros(42)
    for i in range(42):
        d.write(data)
    with pytest.raises(ValueError, match='Dataset is full'):
        d.write(data)


def test_dataset_write_and_get_vectors():
    d = _Dataset(42)
    data = np.random.rand(42, 123)
    for dat in data:
        d.write(dat)
    assert np.array_equal(d.get_data(), data)
    assert d.data is None
    assert d.current_index is None


def test_dataset_write_and_get_scalar():
    d = _Dataset(42)
    data = np.random.rand(42)
    for dat in data:
        d.write(dat)
    got_data = d.get_data()
    assert got_data.shape == (42, 1)
    assert np.array_equal(got_data.squeeze(), data)
    assert d.data is None
    assert d.current_index is None


def test_dataset_write_and_get_not_full():
    d = _Dataset(42)
    data = np.random.rand(33, 123)
    for dat in data:
        d.write(dat)
    assert np.array_equal(d.get_data(), data)
    assert d.data is None
    assert d.current_index is None


####################
#  BufferedWriter  #
####################

def test_buffered_writer_get_reader(output_ets_filename, ets_reader):
    scalar_data = np.arange(len(ets_reader))
    bw = BufferedETSWriter(output_ets_filename)
    bw.write_headers({'foo': 'bar', 'test': 42})
    for i, trace in enumerate(ets_reader):
        trace.scalar = scalar_data[i]
        samples = trace.samples[:]
        bw.write_trace_object_and_points(trace, samples)

    reader = bw.get_reader()
    assert np.array_equal(ets_reader.samples[:], reader.samples[:])
    assert np.array_equal(ets_reader.ciphertext[:], reader.ciphertext[:])
    assert np.array_equal(ets_reader.foo_bar[:], reader.foo_bar[:])
    assert np.array_equal(ets_reader.plaintext[:], reader.plaintext[:])
    assert np.array_equal(scalar_data, reader.scalar[:].squeeze())


def test_buffered_writer_close(output_ets_filename, ets_reader):
    scalar_data = np.arange(len(ets_reader))
    bw = BufferedETSWriter(output_ets_filename, buffer_length=42)
    bw.write_headers({'foo': 'bar', 'test': 42})
    for i, trace in enumerate(ets_reader):
        trace.scalar = scalar_data[i]
        samples = trace.samples[:]
        bw.write_trace_object_and_points(trace, samples)
    bw.close()

    reader = read_ths_from_ets_file(output_ets_filename)
    assert np.array_equal(ets_reader.samples[:], reader.samples[:])
    assert np.array_equal(ets_reader.ciphertext[:], reader.ciphertext[:])
    assert np.array_equal(ets_reader.foo_bar[:], reader.foo_bar[:])
    assert np.array_equal(ets_reader.plaintext[:], reader.plaintext[:])
    assert np.array_equal(scalar_data, reader.scalar[:].squeeze())


def test_buffered_writer_flush_when_empty(output_ets_filename, ets_reader):
    bw = BufferedETSWriter(output_ets_filename, buffer_length=42)
    for trace in ets_reader[:42]:
        samples = trace.samples[:]
        bw.write_trace_object_and_points(trace, samples)
    bw.flush()


def test_buffered_writer_raises_if_one_dataset_has_different_size(output_ets_filename, ets_reader):
    bw = BufferedETSWriter(output_ets_filename, buffer_length=42)
    for trace in ets_reader[:5]:
        samples = trace.samples[:]
        bw.write_trace_object_and_points(trace, samples)
    bw.datasets[_SAMPLES_KEY].write(ets_reader[42].samples[:])
    with pytest.raises(ValueError, match='All Datasets must have the same length'):
        bw.flush()


# https://gitlab.com/eshard/estraces/-/issues/40
def test_buffered_writer_ok_with_strings(output_ets_filename, ets_reader):
    bw = BufferedETSWriter(output_ets_filename, buffer_length=42)
    for trace in ets_reader[:5]:
        trace.foo = 'bar'
        samples = trace.samples[:]
        bw.write_trace_object_and_points(trace, samples)
    bw.flush()



import warnings
import typing


class TraceHeaderSetDeprecated:

    @property
    def metadata_tags(self):
        return list(self.metadatas.keys())

    def get_trace_by_index(self, x):
        warnings.warn(
            "TraceHeaderSet.get_trace_by_index method is deprecated. Use slicing ths[x] instead.",
            DeprecationWarning
        )
        return self[x]

    def get(self, attr, *args, **kwargs):
        warnings.warn(
            "TraceHeaderSet.get method is deprecated. Use .samples attributes or .metadatas attributes instead.",
            DeprecationWarning
        )
        if attr in ('data', 'points'):
            return self.get_points(*args, **kwargs)
        try:
            attribute = getattr(self, attr, *args, **kwargs)
        except AttributeError:
            attribute = getattr(self._reader, attr, *args, **kwargs)
        if isinstance(attribute, typing.Callable):
            return attribute()
        return attribute

    def get_attr(self, attr, *args, **kwargs):
        warnings.warn(
            "TraceHeaderSet.get_attr method is deprecated. Use .samples or .metadatas attributes instead.",
            DeprecationWarning
        )
        return self.get(attr, *args, **kwargs)

    def get_points(self, frame=slice(None, None)):
        warnings.warn(
            "TraceHeaderSet.get_points method is deprecated. Use samples attribute.",
            DeprecationWarning
        )
        if frame is None:
            frame = ...
        if isinstance(frame, range):
            frame = slice(frame.start, frame.stop, frame.step)
        if isinstance(frame, tuple):
            frame = list(frame)
        return self.samples[:, frame]

    def __getattr__(self, name):
        if "get_" in name:
            key = "".join("_".join(name.split("_")[1:]))
            try:
                metadata_value = self.metadatas[key]

                def _():
                    return metadata_value

                warnings.warn(
                    f'TraceHeaderSet.{name} method is deprecated. Use .samples or .metadatas attributes instead.',
                    DeprecationWarning
                )
                return _
            except KeyError:
                raise AttributeError(f'Attribute {name} does not exist on {self}.')
        raise AttributeError(f'Attribute {name} does not exist on {self}.')

    @property
    def h5_file(self):
        try:
            return self._reader._h5_file
        except AttributeError:
            pass

    def close(self):
        try:
            self._reader._h5_file.close()
        except AttributeError:
            pass



from .context import estraces, patch_open, Counter  # noqa
import pytest
import numpy as np
from estraces.formats import bin_extractor, bin_format
import binascii
import cloudpickle


test_traces_fn_pattern = "./tests/samples/AESEncrypt.500MSs.cur.*"
test_trace_fn = "./tests/samples/AESEncrypt.500MSs.cur.00000"
test_trace_fn_pattern = "./tests/samples/AESEncrypt.500MSs.cur.0000*"
aes00000_100 = np.array([78, 82, 83, 87, 87, 88, 86, 90, 87, 87, 84, 89, 86, 86, 86, 88,
                         84, 86, 84, 87, 86, 85, 83, 88, 80, 78, 79, 80, 79, 77, 76, 79,
                         75, 76, 77, 78, 76, 74, 75, 73, 72, 72, 70, 70, 67, 69, 66, 69,
                         67, 73, 76, 80, 81, 87, 89, 94, 95, 100, 102, 105, 103, 105, 107, 105,
                         107, 105, 107, 106, 105, 103, 102, 105, 102, 104, 102, 102, 96, 97, 98, 97,
                         94, 97, 91, 94, 89, 92, 88, 90, 85, 85, 81, 87, 80, 80, 80, 76,
                         77, 74, 72, 70], dtype=np.uint8)
aes00000_last = np.array([49, 51, 56, 60, 63, 67, 73, 79, 80, 82, 85, 88, 89, 91, 90, 90,
                          89, 87, 92, 93, 88, 89, 87, 89, 86, 86, 85, 88, 81, 83, 81, 84,
                          78, 79, 78, 76, 74, 74, 74, 77, 72, 75, 71, 72, 70, 72, 68, 70,
                          67, 68, 66, 66, 61, 64, 61, 65, 61, 67, 69, 70, 71, 71, 75, 75,
                          78, 76, 76, 78, 77, 75, 76, 77, 78, 79, 82, 87, 89, 94, 95, 101,
                          103, 107, 111, 118, 119, 126, 128, 129, 131, 132, 138, 139, 138, 141, 139, 143,
                          144, 142, 145, 147], dtype=np.uint8)


METADATAS = {
    "my_property_2": bin_extractor.PatternExtractor(r"([0-9]{5})", num=0, unhexlify=False),
    "my_property_3": bytes([1, 10, 65]),  # bytes
    "my_property_4": np.array([4, 5, 6]),  # np.ndarray
    "my_property_5": (7, 8, 9),
}


@pytest.fixture
def ths():
    return bin_format.read_ths_from_bin_filenames_pattern(
        filename_pattern=test_traces_fn_pattern,
        dtype="uint8",
        metadatas_parsers=METADATAS,
        offset=0,
    )


@pytest.fixture
def ths_offset():
    return bin_format.read_ths_from_bin_filenames_pattern(
        filename_pattern=test_traces_fn_pattern,
        dtype="uint8",
        metadatas_parsers=METADATAS,
        offset=10,
    )


def test_bin_format_initialization_raises_error_with_improper_dtype():
    with pytest.raises(TypeError):
        bin_format.BinFormat(
            filenames=bin_format.get_sorted_filenames(pattern=test_traces_fn_pattern),
            dtype="dfjdlkfjdkljl",
            metadatas_parsers=METADATAS,
        )


def test_bin_format_initialization_raises_error_with_improper_filenames():
    with pytest.raises(TypeError):
        bin_format.BinFormat(
            filenames="ruieuirur", dtype="uint8", metadatas_parsers=METADATAS
        )


def test_empty_bin_format_returns_no_data():
    ths = bin_format.read_ths_from_bin_filenames_list(filenames_list=[], dtype='uint8', offset=0, metadatas_parsers={'meta1': lambda t: t})
    assert 0 == len(ths)
    assert (0,) == ths.samples.shape
    assert [('meta1', [])] == [(k, v.tolist()) for k, v in ths.metadatas.items()]


def test_bin_format_initialization():
    bin_engine = bin_format.BinFormat(
        filenames=bin_format.get_sorted_filenames(pattern=test_traces_fn_pattern),
        dtype="uint8",
        metadatas_parsers=METADATAS,
    )
    assert isinstance(bin_engine, bin_format.BinFormat)


def test_initialize_with_factory_function(ths):
    assert isinstance(ths, estraces.TraceHeaderSet)
    assert len(ths) == 100


def test_traces_attribute(ths):
    assert isinstance(ths.traces[0], estraces.Trace)
    assert len(ths.traces) == 100


def test_samples_slicing(ths, ths_offset):
    files = bin_format.get_sorted_filenames(pattern=test_traces_fn_pattern)
    for t in [ths, ths_offset]:
        for index, trc in enumerate(t.traces):
            filename = files[index]
            with open(filename, "rb") as fid:
                fid.seek(t._reader._offset)
                exp_data = fid.read()
                exp_data = np.frombuffer(exp_data, dtype="uint8")
                assert np.array_equal(trc.samples[:], exp_data)

    # Get with a frame
    for t in [ths, ths_offset]:
        for index, trc in enumerate(t.traces):
            filename = files[index]
            with open(filename, "rb") as fid:
                fid.seek(10 + t._reader._offset)
                exp_data = fid.read(40)
                exp_data = np.frombuffer(exp_data, dtype="uint8")
                assert np.array_equal(trc.samples[10:50], exp_data)


def test_samples_slicing_with_zero_length_slices(ths):
    assert isinstance(ths.samples[:, 100:100], np.ndarray)
    assert 0 == ths.samples[:, 100:100].shape[1]

    assert isinstance(ths.samples[100:100, :], np.ndarray)
    assert 0 == ths.samples[100:100, :].shape[0]

    assert isinstance(ths[0].samples[100:100], np.ndarray)
    assert 0 == ths.samples[100:100].shape[0]


def test_samples_slicing_with_negative_slices(ths):
    le = len(ths[0].samples)
    assert ths[0].samples[le - 200:le - 100].tolist() == ths[0].samples[-200: -100].tolist()
    assert [] == ths[0].samples[-100: -200].tolist()

    le = ths.samples.shape[1]
    assert ths.samples[:, le - 200: le - 100].tolist() == ths.samples[:, -200: -100].tolist()
    assert [[] for i in range(ths.samples.shape[0])] == ths.samples[:, -100: -200].tolist()

    le = ths.samples.shape[0]
    assert ths.samples[le - 2:le - 1, :].tolist() == ths.samples[-2: -1, :].tolist()
    assert [] == ths.samples[-1:-2, :].tolist()


def test_ths_slicing(ths):

    assert isinstance(ths[2], estraces.Trace)
    with pytest.raises(IndexError):
        ths[12:12, 0]
        ths["jfkj"]

    ths2 = ths[10:15]
    ths3 = ths[[1, 3]]
    ths4 = ths[:15]
    ths5 = ths[5:]
    assert isinstance(ths2, estraces.TraceHeaderSet)
    assert isinstance(ths3, estraces.TraceHeaderSet)
    assert isinstance(ths4, estraces.TraceHeaderSet)
    assert isinstance(ths5, estraces.TraceHeaderSet)

    assert len(ths2) == 5
    assert len(ths3) == 2
    assert len(ths4) == 15
    assert len(ths5) == len(ths) - 5

    assert id(ths) != id(ths2)

    origin_traces = ths.traces[10:15]
    for idx, trc in enumerate(ths2.traces):
        origin_trc = origin_traces[idx]
        assert id(trc) != id(origin_trc)
        assert trc.samples.tolist() == origin_trc.samples.tolist()
        for k, v in trc.metadatas.items():
            if isinstance(v, np.ndarray):
                assert np.array_equal(v, origin_trc.metadatas[k])
            else:
                assert v == origin_trc.metadatas[k]

    for idx, trc in enumerate(ths3.traces):
        o_ind = 1 if idx == 0 else 3
        origin_trc = ths.traces[o_ind]
        assert id(trc) != id(origin_trc)
        assert trc.samples.tolist() == origin_trc.samples.tolist()
        for k, v in trc.metadatas.items():
            if isinstance(v, np.ndarray):
                assert np.array_equal(v, origin_trc.metadatas[k])
            else:
                assert v == origin_trc.metadatas[k]


def test_bin_format_samples_api(ths, ths_offset):
    data = ths.samples[0, :]
    assert np.array_equal(data[0:100], aes00000_100)

    data = ths.samples[0, 10:20]
    assert data.ndim == 1
    assert np.array_equal(data, aes00000_100[10:20])

    data = ths.samples[0, [1, 3, 4, 9]]
    assert data.ndim == 1
    assert np.array_equal(data, aes00000_100[[1, 3, 4, 9]])

    # With a too large slice via indexing
    data = ths.samples[0, 99900:200000]
    assert np.array_equal(data, aes00000_last)

    # Requesting a single point via indexing (getitem)
    data = ths.samples[0, 7]
    assert np.array_equal(data, aes00000_100[7])

    # Requesting a too large single point via indexing (getitem)
    with pytest.raises(IndexError):
        ths.samples[0, 700000]

    # With a wrong index type
    with pytest.raises(IndexError):
        ths.samples[0, "abcd"]

    # With an offset
    data = ths_offset.samples[0]
    assert len(data) == 99990
    assert np.array_equal(data[0:90], aes00000_100[10:])

    trc = ths[0]
    assert aes00000_100[12] == trc.samples[12]

    assert isinstance(trc.samples[12], np.uint8)

    assert np.array_equal(
        aes00000_100[1:10:2],
        trc.samples[1:10:2]
    )


def test_metadatas(ths):

    prop2 = ths.my_property_2
    prop3 = ths.my_property_3
    prop4 = ths.my_property_4
    # Get a wrong attribute
    with pytest.raises(AttributeError):
        ths.fake

    for trc_idx in range(len(ths)):
        assert "{:05d}".format(trc_idx) == prop2[trc_idx]
        assert bytes([1, 10, 65]) == prop3[trc_idx]
        assert np.array_equal(np.array([4, 5, 6]), prop4[trc_idx])

    assert isinstance(ths.metadatas, estraces.traces.metadatas.Metadatas)
    assert ths.metadatas["my_property_2"].tolist() == ths.my_property_2.tolist()
    assert ths.metadatas["my_property_3"].tolist() == ths.my_property_3.tolist()


def test_bin_format_fetch_samples_always_returns_2d_array(ths):
    assert 2 == ths._reader.fetch_samples(traces=[1, 2, 3], frame=...).ndim
    assert 2 == ths._reader.fetch_samples(traces=range(len(ths._reader)), frame=1).ndim
    assert 2 == ths._reader.fetch_samples(traces=[1], frame=slice(10, 20, 1)).ndim
    assert 2 == ths._reader.fetch_samples(traces=[1], frame=slice(10, 10, 1)).ndim
    assert 2 == ths._reader.fetch_samples(traces=[1], frame=1).ndim
    assert 2 == ths._reader.fetch_samples(traces=[1, 4], frame=1).ndim


def test_direct_value():
    """Verify that we handle any kind of input from the user for plain, cipher and key."""
    expect = b'\x03\xd20\xbf\x92o\xd8\xf5'
    in_1 = "03D230BF926FD8F5"
    in_2 = [3, 210, 48, 191, 146, 111, 216, 245]

    out_1 = bin_extractor.DirectValue(in_1)
    assert expect == out_1.get_text(None)
    out_2 = bin_extractor.DirectValue(in_2)
    assert expect == out_2.get_text(None)


def test_direct_value_wrong_inputs():
    """Verify that we handle any kind of input from the user for plain, cipher and key."""
    in_1 = 0x03D230BF926FD8F5
    in_2 = "0x03D230BF926FD8F5"
    in_3 = "03D230BF926FD8F57"
    in_4 = [0, 1, 255, 257]

    with pytest.raises(ValueError) as exc:
        bin_extractor.DirectValue(in_1)
    assert str(exc.value) == "Invalid value format <class 'int'>. HexString without the '0x' prefix or a list of integer in range(0, 256) is expected."

    with pytest.raises(ValueError) as exc:
        bin_extractor.DirectValue(in_2)
    assert str(exc.value) == "Invalid character in input string."
    with pytest.raises(ValueError) as exc:
        bin_extractor.DirectValue(in_3)
    assert str(exc.value) == "Odd-length string."
    with pytest.raises(ValueError) as exc:
        bin_extractor.DirectValue(in_4)
    assert str(exc.value) == "bytes must be in range(0, 256)"


def test_header_extractor():
    filename = "tests/samples/DESi386__03D230BF926FD8F5_DB0F84C83D856553.bin"
    key = bin_extractor.HeaderExtractor(start=0, count=8)
    cipher = bin_extractor.HeaderExtractor(start=8, count=8)
    k = key.get_text(filename)
    c = cipher.get_text(filename)

    assert k == binascii.unhexlify('1c0000001cec7ff6')
    assert c == binascii.unhexlify('50cc7ef600e07ff6')


def test_header_extractor_wrong_inputs():
    # No arguments
    with pytest.raises(ValueError) as exc:
        bin_extractor.HeaderExtractor()
    assert str(exc.value) == "No 'start' offset specified."

    # No end or count specified
    with pytest.raises(ValueError) as exc:
        bin_extractor.HeaderExtractor(start=0)
    assert str(exc.value) == "No 'end' offset or 'count' specified."

    # Both end AND count specified
    with pytest.raises(ValueError) as exc:
        bin_extractor.HeaderExtractor(start=0, count=10, end=12)
    assert str(exc.value) == "Use either 'end' or 'count' option, not both."

    # Wrong start value
    with pytest.raises(ValueError) as exc0:
        bin_extractor.HeaderExtractor(start=-332, count=10)
    with pytest.raises(ValueError) as exc1:
        bin_extractor.HeaderExtractor(start=1.2, count=10)
    with pytest.raises(ValueError) as exc2:
        bin_extractor.HeaderExtractor(start=1, count=-2)
    with pytest.raises(ValueError) as exc3:
        bin_extractor.HeaderExtractor(start=1, count=1.2)
    with pytest.raises(ValueError) as exc4:
        bin_extractor.HeaderExtractor(start=1, end=-2)
    with pytest.raises(ValueError) as exc5:
        bin_extractor.HeaderExtractor(start=1, end=1.2)

    assert 'must be positive integer' in str(exc0.value)
    assert 'must be positive integer' in str(exc1.value)
    assert 'must be positive integer' in str(exc2.value)
    assert 'must be positive integer' in str(exc3.value)
    assert 'must be positive integer' in str(exc4.value)
    assert 'must be positive integer' in str(exc5.value)

    # Incoherent start, end values
    with pytest.raises(ValueError) as exc0:
        bin_extractor.HeaderExtractor(start=0, count=0)
    with pytest.raises(ValueError) as exc1:
        bin_extractor.HeaderExtractor(start=0, end=0)
    assert str(exc0.value) == 'The end offset must be greater than the start offset.'
    assert str(exc1.value) == 'The end offset must be greater than the start offset.'

    # Specified both cound and end arguments
    with pytest.raises(ValueError) as exc0:
        bin_extractor.HeaderExtractor(start=0, end=1, count=1)
    with pytest.raises(ValueError) as exc1:
        bin_extractor.HeaderExtractor(start=0, end=0, count=0)
    assert str(exc0.value) == "Use either 'end' or 'count' option, not both."
    assert str(exc1.value) == "Use either 'end' or 'count' option, not both."


def test_pattern_extractor():
    filename = "wave_aist-aes-agilent_2009-12-30_22-53-19_k=0000000000000003243f6a8885a308d3_m=0000000000000002b7e151628aed2a6a_c=e6a636e30c85f35e980f3546a04daff7.bin"  # noqa
    key = bin_extractor.PatternExtractor(pattern=r"([A-Fa-f0-9]{32})", num=0)
    msg = bin_extractor.PatternExtractor(pattern=r"([A-Fa-f0-9]{32})", num=1)
    cipher = bin_extractor.PatternExtractor(pattern=r"([A-Fa-f0-9]{32})", num=2)
    test_0 = bin_extractor.PatternExtractor(pattern=r"([A-Fa-f0-9]{16})")
    test_1 = bin_extractor.PatternExtractor(pattern=r"([A-Fa-f0-9]{16})", num=0)
    test_2 = bin_extractor.PatternExtractor(pattern=r"([A-Fa-f0-9]{16})", num=1)
    test_3 = bin_extractor.PatternExtractor(pattern=r"([A-Fa-f0-9]{35})")

    k = key.get_text(filename)
    m = msg.get_text(filename)
    c = cipher.get_text(filename)

    t0 = test_0.get_text(filename)
    t1 = test_1.get_text(filename)
    t2 = test_2.get_text(filename)

    assert k == binascii.unhexlify('0000000000000003243f6a8885a308d3')
    assert m == binascii.unhexlify('0000000000000002b7e151628aed2a6a')
    assert c == binascii.unhexlify('e6a636e30c85f35e980f3546a04daff7')
    assert t0 == binascii.unhexlify('0000000000000003')
    assert t1 == binascii.unhexlify('0000000000000003')
    assert t2 == binascii.unhexlify('243f6a8885a308d3')

    with pytest.raises(ValueError) as exc:
        test_3.get_text(filename)
    assert str(exc.value) == "Pattern '([A-Fa-f0-9]{35})' not found in 'wave_aist-aes-agilent_2009-12-30_22-53-19_k=0000000000000003243f6a8885a308d3_m=0000000000000002b7e151628aed2a6a_c=e6a636e30c85f35e980f3546a04daff7.bin'."  # noqa


def test_file_pattern_extractor():
    """Verify that the 5 first messages are what we expect."""
    expected = ['D9EDDB17D7B1B4F406685DA69CA71DA1',
                '8384F00F9869555D3D1401E9679A4A6B',
                '105FDD49C7EF0961BB993263B48ECF34',
                '9A870CAC7B9C38E99A1E88DA9BAE02D4',
                'E87303E1BE987A9CE1FB6B16F460AD14']
    msg = []
    tmp = bin_extractor.FilePatternExtractor(
        "tests/samples/plain.txt", pattern=r"([A-Fa-f0-9]{32})", replace=r"\1")
    for _ in range(5):
        msg.append(tmp.get_text(None))

    for it in enumerate(expected):
        m = binascii.unhexlify(it[1])
        assert m == msg[it[0]]


def test_file_pattern_extractor_wrong_inputs():
    """Verify that error are raised with wrong user inputs."""
    # Wrong file path
    with pytest.raises(FileNotFoundError) as exc:
        bin_extractor.FilePatternExtractor("tests/samples/plainFAKE.txt", pattern=r"([A-Fa-f0-9]{32})", replace=r"\1")
    assert str(exc.value) == "'tests/samples/plainFAKE.txt' not found"


def test_file_pattern_extractor_long():
    # Extract the meta data from the same ascii files
    file = 'tests/samples/metadatas.txt'
    key = bin_extractor.FilePatternExtractor(file, r"([A-Fa-f0-9]{32})", num=0)
    plain = bin_extractor.FilePatternExtractor(file, r"([A-Fa-f0-9]{32})", num=1)
    cipher = bin_extractor.FilePatternExtractor(file, r"([A-Fa-f0-9]{32})", num=2)
    mask1 = bin_extractor.FilePatternExtractor(file, r"([A-Fa-f0-9]{32})", num=3)
    mask2 = bin_extractor.FilePatternExtractor(file, r"([A-Fa-f0-9]{32})", num=4)
    mask3 = bin_extractor.FilePatternExtractor(file, r"([A-Fa-f0-9]{32})", num=5)

    ths = bin_format.read_ths_from_bin_filenames_pattern(
        filename_pattern=test_traces_fn_pattern,
        dtype='uint8',
        offset=0,
        metadatas_parsers={
            'key': key,
            'plain': plain,
            'cipher': cipher,
            'mask1': mask1,
            'mask2': mask2,
            'mask3': mask3
        }
    )

    with patch_open(Counter()) as counter:
        ths.metadatas['plain']
        ths.metadatas['mask3']
        ths.metadatas['cipher']
        ths.metadatas['key']
        ths.metadatas['mask1']
        ths.metadatas['mask2']
        ths.plain
        ths.mask3
        ths.cipher
        ths.key
        ths.mask1
        ths.mask2
        trc = ths[0]
        trc.plain
        trc.mask3
        trc.cipher
        trc.key
        trc.mask1
        trc.mask2
        for v in counter.counts.values():
            assert v < 7  # One file open for each meta data

    with patch_open(Counter()) as counter:
        ths.metadatas['plain']
        ths_2 = ths[0:50]
        ths_2.plain
        for v in counter.counts.values():
            assert v < 2


@pytest.fixture(params=[bin_format.BinFormat, bin_format.read_ths_from_bin_filenames_list])
def bin_factory(request):
    return request.param


def test_bin_reader_raises_exception_if_invalid_padding_mode(bin_factory):
    with pytest.raises(AttributeError):
        bin_factory([test_trace_fn], dtype='uint8', metadatas_parsers={}, offset=0, padding_mode='TRU')


def test_ths_bin_pattern_factory_raises_exception_if_invalid_padding_mode():
    with pytest.raises(AttributeError):
        bin_format.read_ths_from_bin_filenames_pattern('./tests/samples/*difflength.bin', dtype='uint8', metadatas_parsers={}, padding_mode='TRU')


def test_default_bin_reader_raise_exception_at_init_with_padding_mode_default_and_traces_not_of_the_same_size(bin_factory):
    filenames = bin_format.get_sorted_filenames('./tests/samples/*difflength.bin')
    with pytest.raises(ValueError):
        bin_factory(filenames, dtype='uint8', metadatas_parsers={})


def test_default_pattern_ths_factory_raise_exception_at_init_with_padding_mode_default_and_traces_not_of_the_same_size(bin_factory):
    with pytest.raises(ValueError):
        bin_format.read_ths_from_bin_filenames_pattern('./tests/samples/*difflength.bin', dtype='uint8', metadatas_parsers={})


def test_bin_reader_with_truncate_padding_mode_truncate_too_long_traces():
    filenames = bin_format.get_sorted_filenames('./tests/samples/*difflength.bin')
    binf = bin_format.BinFormat(filenames, 'uint8', {}, padding_mode=bin_format.PaddingMode.TRUNCATE)
    for i in range(3):
        assert 990 == binf.get_trace_size(i)
        assert 990 == binf.fetch_samples(traces=[i], frame=None).shape[1]
    assert 2 == len(binf[:2])


def test_bin_reader_with_pad_padding_mode_pad_too_short_traces():
    filenames = bin_format.get_sorted_filenames('./tests/samples/*difflength.bin')
    binf = bin_format.BinFormat(filenames, 'uint8', {}, padding_mode=bin_format.PaddingMode.PAD)
    for i in range(3):
        assert 1010 == binf.get_trace_size(i)
        assert 1010 == binf.fetch_samples(traces=[i], frame=None).shape[1]
    assert [0] * 10 == binf.fetch_samples(traces=[0], frame=None)[0, -10:].tolist()
    assert [0] * 20 == binf.fetch_samples(traces=[2], frame=None)[0, -20:].tolist()
    assert 2 == len(binf[:2])


def test_serialization(ths):
    dumped = cloudpickle.dumps(ths)
    loaded_ths = cloudpickle.loads(dumped)
    assert np.array_equal(loaded_ths.samples[:], ths.samples[:])
    assert np.array_equal(loaded_ths.my_property_2[:], ths.my_property_2[:])
    assert np.array_equal(loaded_ths.my_property_3[:], ths.my_property_3[:])
    assert np.array_equal(loaded_ths.my_property_4[:], ths.my_property_4[:])
    assert np.array_equal(loaded_ths.my_property_5[:], ths.my_property_5[:])


# see https://gitlab.com/eshard/estraces/-/issues/45
def test_filename_property(ths):
    assert np.array_equal(ths.filename, ths._reader._filenames)



from .context import estraces  # noqa
import pytest
import numpy as np
from .conftest import PLAINS, CIPHERS, INDICES


@pytest.fixture(params=['1 trace', 'all traces'])
def metadatas(request, fmt):
    return estraces.traces.metadatas.Metadatas(reader=fmt, trace_id=0 if request.param == '1 trace' else None)


@pytest.fixture
def metas_1_trace(fmt):
    return estraces.traces.metadatas.Metadatas(reader=fmt, trace_id=0)


@pytest.fixture
def metas_1_ths(fmt):
    return estraces.traces.metadatas.Metadatas(reader=fmt)


def test_metadatas_init_with_improper_reader_type_raises_exception():
    with pytest.raises(TypeError):
        estraces.traces.metadatas.Metadatas(reader="oups")


def test_metadatas_repr(metadatas):
    assert repr(metadatas)


def test_read_metadatas_key_throws_key_error_on_missing_key(metadatas):
    with pytest.raises(KeyError):
        metadatas["missing"]
    with pytest.raises(KeyError):
        metadatas[3]


def test_read_metadatas_key_throws_type_error_on_incorrect_key_type(metadatas):
    with pytest.raises(TypeError):
        metadatas[np.array(range(2))]
    with pytest.raises(TypeError):
        metadatas[np.array([1, 2, 3], dtype="uint8")]


def test_read_metadata_returns_array_of_values(metas_1_ths):
    assert isinstance(metas_1_ths["ciphertext"], np.ndarray)
    assert np.array_equal(metas_1_ths["ciphertext"], CIPHERS)


def test_read_metadata_returns_a_mapping(metadatas):
    assert [k for k in sorted(metadatas.keys())] == ["ciphertext", "indices", "plain_t", "plaintext"]
    if metadatas._trace_id is None:
        for k, v in metadatas.items():
            assert k in ("plaintext", "ciphertext", "indices", "plain_t")
            if k == 'ciphertext':
                assert np.array_equal(v, CIPHERS)
            elif k == 'plaintext' or k == 'plain_t':
                assert np.array_equal(v, PLAINS)
            else:
                assert np.array_equal(v, INDICES)
    else:
        for k, v in metadatas.items():
            assert k in ("plaintext", "ciphertext", "indices", "plain_t")
            if k == 'ciphertext':
                assert np.array_equal(v, CIPHERS[0])
            elif k == 'plaintext' or k == 'plain_t':
                assert np.array_equal(v, PLAINS[0])
            else:
                assert v == INDICES[0]


def test_read_metadata_return_values(metas_1_trace):
    assert isinstance(metas_1_trace["ciphertext"], np.ndarray)
    assert np.array_equal(metas_1_trace["ciphertext"], CIPHERS[0])
    assert isinstance(metas_1_trace["indices"], str)
    assert metas_1_trace["indices"] == INDICES[0]


def test_metadatas_cache(metas_1_ths):
    assert metas_1_ths._cache == {}
    metas_1_ths["ciphertext"]
    assert "ciphertext" in metas_1_ths._cache
    assert np.array_equal(metas_1_ths._cache['ciphertext'], CIPHERS)


def test_metadatas_set_new_attribute(metadatas):
    metadatas['new_attribute'] = 'foo'

    assert metadatas['new_attribute'] == 'foo'



from .trace import Trace
from .trace_header_set import TraceHeaderSet, build_trace_header_set  # noqa : C901
from .samples import Samples  # noqa : C901
from .metadatas import Metadatas
from .abstract_reader import AbstractReader
from . import headers  # noqa : C901

__all__ = [
    "Trace",
    "TraceHeaderSet",
    "build_trace_header_set"
    "Samples",
    "Metadatas",
    "AbstractReader"
]



from .context import estraces # noqa
import pytest
import numpy as np
from estraces.formats import ets_format, ram_format
import cloudpickle

path = 'tests/samples/test.ets'


def load_data():
    ets = ets_format.read_ths_from_ets_file(filename=path)
    samples = ets.samples[:].copy()
    plaintext = ets.plaintext[:].copy()
    foo_bar = ets.foo_bar[:].copy()
    ciphertext = ets.ciphertext[:].copy()
    return samples, plaintext, foo_bar, ciphertext


@pytest.fixture
def ram1():
    samples, plaintext, foo_bar, ciphertext = load_data()
    return ram_format.read_ths_from_ram(samples=samples, plaintext=plaintext, foo_bar=foo_bar, ciphertext=ciphertext)


def test_instanciate(ram1):
    assert isinstance(ram1, estraces.TraceHeaderSet)
    assert 200 == len(ram1)


def test_raise_type_error_if_inputs_are_not_ndarray():
    samples, plaintext, foo_bar, ciphertext = load_data()
    with pytest.raises(TypeError):
        ram_format.read_ths_from_ram(samples.tolist(), plaintext=plaintext, foobar=foo_bar, ciphertext=ciphertext)
    with pytest.raises(TypeError):
        ram_format.read_ths_from_ram(samples, plaintext=plaintext.tolist(), foo_bar=foo_bar, ciphertext=ciphertext)


def test_raise_type_error_if_inputs_are_not_correct_arrays():
    _, plaintext, foo_bar, ciphertext = load_data()
    with pytest.raises(TypeError):
        ram_format.read_ths_from_ram(np.arange(200), plaintext=plaintext, foobar=foo_bar, ciphertext=ciphertext)


def test_incompatible_shapes_raise_value_error():
    samples, plaintext, foo_bar, ciphertext = load_data()
    with pytest.raises(ValueError):
        ram_format.read_ths_from_ram(samples[:10], plaintext=plaintext, foobar=foo_bar, ciphertext=ciphertext)
    with pytest.raises(ValueError):
        ram_format.read_ths_from_ram(samples, plaintext=plaintext[:10], foo_bar=foo_bar, ciphertext=ciphertext)
    with pytest.raises(ValueError):
        ram_format.read_ths_from_ram(samples, plaintext=np.arange(300), foo_bar=foo_bar, ciphertext=ciphertext)


def test_read_metadatas(ram1):
    assert isinstance(ram1.metadatas, estraces.traces.metadatas.Metadatas)
    for v in ram1.metadatas.keys():
        assert v in ['ciphertext', 'foo_bar', 'plaintext']
    assert [2, 9, 15, 8, 11, 8, 2, 13] == ram1.metadatas['plaintext'][0].tolist()
    assert [2, 9, 15, 8, 11, 8, 2, 13] == ram1.plaintext[0].tolist()
    assert [2, 9, 15, 8, 11, 8, 2, 13] == ram1[0].plaintext.tolist()
    for _, v in ram1.metadatas.items():
        assert 2 == v.ndim
    for _, v in ram1[0].metadatas.items():
        assert 1 == v.ndim
    assert [
        [3, 10, 16, 9, 12, 9, 3, 14],
        [4, 11, 17, 10, 13, 10, 4, 15],
        [5, 12, 18, 11, 14, 11, 5, 16]
    ] == ram1.plaintext[1: 4].tolist()


def test_read_samples(ram1):
    assert isinstance(ram1.samples, estraces.traces.samples.Samples)
    assert ram1.samples.ndim == 2
    assert isinstance(ram1[0].samples, estraces.traces.samples.Samples)
    assert ram1[0].samples.ndim == 1
    assert [
        [2, 11, 21],
        [4, 13, 23],
        [6, 15, 25]
    ] == ram1.samples[1: 6: 2, [1, 10, 20]].tolist()

    assert [
        [2, 11, 21],
        [4, 13, 23],
        [6, 15, 25]
    ] == ram1.samples[1: 6: 2, np.array([1, 10, 20])].tolist()

    assert [
        [2, 3, 4, 5, 6, 7, 8, 9, 10],
        [4, 5, 6, 7, 8, 9, 10, 11, 12],
        [6, 7, 8, 9, 10, 11, 12, 13, 14]
    ] == ram1.samples[1: 6: 2, 1:10].tolist()

    # Manage limits
    assert 0 == ram1.samples[0, 0]
    assert [1, 2, 3, 4, 5, 6, 7, 8, 9] == ram1.samples[0, 1:10].tolist()
    assert [1, 2, 3, 4, 5, 6, 7, 8, 9] == ram1.samples[1:10, 0].tolist()

    assert [1, 3, 4, 7, 9] == ram1[0].samples[np.array([1, 3, 4, 7, 9])].tolist()

    assert [1, 4, 3, 7, 4, 9] == ram1[0].samples[np.array([1, 4, 3, 7, 4, 9])].tolist()

    assert list(range(33)) == ram1[0].samples[...].tolist()
    assert 12 == ram1[0].samples[12]


def test_slice_set(ram1):
    sub1 = ram1[2: 7]
    sub2 = ram1[[1, 10, 12]]
    assert isinstance(sub1, estraces.TraceHeaderSet)
    assert isinstance(sub2, estraces.TraceHeaderSet)
    assert 5 == len(sub1)
    assert 3 == len(sub2)
    assert [2, 3, 4, 5] == sub1.samples[0, 0:4].tolist()
    assert [6, 7, 8, 9] == sub1.samples[4, 0:4].tolist()
    assert [1, 2, 3, 4] == sub2.samples[0, 0:4].tolist()
    assert [10, 11, 12, 13] == sub2.samples[1, 0:4].tolist()
    assert [12, 13, 14, 15] == sub2.samples[2, 0:4].tolist()
    assert sub1.plaintext.shape[0] == 5

    sub_3 = ram1[0:1]
    assert sub_3.plaintext.shape[0] == 1


def test_split_preserves_length_consistency_of_sub_ths(ram1):
    parts = ram1.split(100)
    for _, p in enumerate(parts):
        assert len(p) in (100, 56)
        assert len(p) == p.samples[:].shape[0]


def test_serialization(ram1):
    ths = ram1
    dumped = cloudpickle.dumps(ths)
    loaded_ths = cloudpickle.loads(dumped)
    assert np.array_equal(loaded_ths.samples[:], ths.samples[:])
    assert np.array_equal(loaded_ths.plaintext[:], ths.plaintext[:])
    assert np.array_equal(loaded_ths.foo_bar[:], ths.foo_bar[:])
    assert np.array_equal(loaded_ths.ciphertext[:], ths.ciphertext[:])


# see https://gitlab.com/eshard/estraces/-/issues/45
def test_filename_property(ram1):
    assert ram1.filename is None



from .context import estraces  # noqa
import pytest
import numpy as np
from estraces import build_trace_header_set
from estraces._legacy_formats.bin_format import BinReader
from estraces._legacy_formats.ets_format import ETSReader


@pytest.fixture
def ths(fmt):
    return build_trace_header_set(reader=fmt, name="Test legacy ths")


def test_init_legacy_ths_raise_deprecation_warnings(ths):
    import warnings
    warnings.simplefilter('default', DeprecationWarning)
    with pytest.warns(DeprecationWarning):
        ths.get_points()


def test_ths_get_points(ths):
    assert np.array_equal(ths.get_points(), ths.samples[:, ...])

    # frame can be range, tuple
    assert np.array_equal(
        ths.samples[:, slice(2, 6, 2)],
        ths.get_points(frame=range(2, 6, 2))
    )

    assert np.array_equal(
        ths.samples[:, [2, 6, 8]],
        ths.get_points(frame=(2, 6, 8))
    )


def test_getters(ths):

    assert np.array_equal(
        ths.plain_t,
        ths.get_plain_t()
    )

    assert np.array_equal(
        ths.plaintext,
        ths.get_plaintext()
    )

    assert np.array_equal(
        ths.ciphertext,
        ths.get_ciphertext()
    )

    assert np.array_equal(
        ths.indices,
        ths.get_indices()
    )


def test_ths_get_method(ths):

    assert np.array_equal(
        ths.plaintext,
        ths.get('plaintext')
    )

    assert ths.samples.tolist() == ths.get('data').tolist()

    assert np.array_equal(
        ths.samples[:, [1, 5]],
        ths.get('points', frame=[1, 5])
    )

    with pytest.raises(AttributeError):
        ths.get('noneattr')


def test_get_attr_method(ths):

    assert np.array_equal(
        ths.plaintext,
        ths.get_attr('plaintext')
    )

    assert ths.samples.tolist() == ths.get_attr('data').tolist()

    assert np.array_equal(
        ths.samples[:, [1, 5]],
        ths.get_attr('points', frame=[1, 5])
    )

    with pytest.raises(AttributeError):
        ths.get_attr('noneattr')


def test_get_trace_by_index(ths):

    assert ths[4] == ths.get_trace_by_index(x=4)

    with pytest.raises(IndexError):
        ths.get_trace_by_index(x=444)


def test_trace_points_method(ths):
    trc = ths[0]
    assert np.array_equiv(trc.samples[:], trc.points())

    # frame can be range, tuple, int
    assert np.array_equal(
        trc.samples[slice(2, 6, 2)],
        trc.points(frame=range(2, 6, 2))
    )

    assert np.array_equal(
        trc.samples[[2, 6, 8]],
        trc.points(frame=(2, 6, 8))
    )

    assert trc.samples[12] == trc.points(frame=12)
    assert isinstance(trc.points(frame=12), np.uint8)


def test_trace_nb_point(ths):
    assert ths[0].nb_point == len(ths[0])


def test_bin_format_ths_filename_attribute():
    test_traces_fn_pattern = "./tests/samples/AESEncrypt.500MSs.cur.*"
    ths = BinReader(filename_pattern=test_traces_fn_pattern, dtype='uint8')
    assert ths[0].filename == ths[0]._reader._filenames[0]
    assert ths.get('filename').tolist() == ths._reader._filenames.tolist()


def test_legacy_ets_initialization():
    ets = ETSReader(filename='tests/samples/test.ets')
    assert ets.samples.tolist() == ets.get_points().tolist()


def test_get_trace_by_index_method(ths):
    assert ths.get_trace_by_index(2) == ths[2]


def test_get_writable_attributes(ths):
    assert list(ths[0].metadatas.keys()) == ths[0].__get_writable_attributes__()
    ths[0].new_prop = 'value'
    exp = [k for k in ths[0].metadatas.keys()]
    exp.append('new_prop')
    assert exp == ths[0].__get_writable_attributes__()


def test_h5_file_property(ths):
    try:
        assert ths.h5_file == ths._reader._h5_file
    except AttributeError:
        assert ths.h5_file is None


def test_h5_close_method(ths):
    assert ths.close() is None



from functools import lru_cache

from ..traces.abstract_reader import AbstractReader
from ..traces.trace_header_set import build_trace_header_set
import estraces

import numpy as _np


def read_ths_from_multiple_ths(*args, **kwargs):
    """Build and return a :class:`TraceHeaderSet` instance as the concatenation of all input THSs.

    Args:
        args: some `TraceHeaderSet` objects.
        kwargs: some `TraceHeaderSet` objects.

    You can pass a headers dict argument to override inconsistent headers between trace header sets or add custom values.

    Note:
        shape and metadata must be consistent between each ths.

    Returns:
        (:obj:`TraceHeaderSet`)

    """
    return build_trace_header_set(
        reader=ConcatFormatReader(*args, **kwargs),
        name='Concat Format THS'
    )


class ConcatFormatReader(AbstractReader):

    def __init__(self, *args, **kwargs):

        headers = kwargs.pop('headers', {})
        check = kwargs.pop('check', True)
        all_args = dict(**{f'ths_{i}': arg for i, arg in enumerate(args)}, **kwargs)

        if check:
            self._check_are_trace_header_sets(all_args)
            self._check_metadata_consistency(all_args)
            self._check_sample_shapes_and_dtype_consistency(all_args)
            self._check_metadata_shapes_and_dtype_consistency(all_args)
            self._headers = self._check_headers(all_args, headers)
        else:
            self._headers = headers

        self.ths_list = list(all_args.values())
        self._ths_dict = all_args
        self._trace_size = self.ths_list[0]._reader.get_trace_size(0)
        self._trace_dtype = self.ths_list[0].samples[0].dtype
        self._metas_infos = self._get_metadatas_infos()

        self._sizes = [len(ths) for ths in self.ths_list]
        self._size = _np.sum(self._sizes)

        self._sub_traceset_indices = _cached_arange(self._size)

    @staticmethod
    def _check_are_trace_header_sets(thss_dict):
        for key in thss_dict.keys():
            if not isinstance(thss_dict[key], estraces.TraceHeaderSet):
                raise TypeError(f'{key} argument must be a `TraceHeaderSet` object, not `{type(thss_dict[key])}`.')

    @staticmethod
    def _check_metadata_consistency(thss_dict):
        key0 = list(thss_dict.keys())[0]
        set0 = set(thss_dict[key0]._reader.metadatas_keys)
        for key in thss_dict.keys():
            set1 = set(thss_dict[key]._reader.metadatas_keys)
            if set0 != set1:
                raise ValueError(f'Inconsistent metadata keys between: {key0}, {set0} and {key}, {set1}')

    @staticmethod
    def _check_headers(thss_dict, headers={}):
        ths_0 = list(thss_dict.values())[0]
        set_0 = set(ths_0.headers.keys())
        for key, ths in thss_dict.items():
            if ths == ths_0:
                continue
            set_1 = set(ths.headers.keys())
            if set_0 != set_1:
                diff = set_0.difference(set_1)
                for key in diff:
                    if key not in headers:
                        raise ValueError(
                            f'Inconsistent headers, {key} missing from one ths.\
                            You should override headers values by passing a headers dict to the constructor.'
                        )
            for k in set_0:
                if k not in headers:
                    val_1 = ths.headers[k]
                    val_2 = ths_0.headers[k]
                    equal = _np.array_equal(val_1, val_2) if isinstance(val_1, _np.ndarray) else val_1 == val_2
                    if not equal:
                        raise ValueError(
                            f'Inconsistent headers values between {ths_0} and {ths} for header {k}.\
                            You should override headers values by passing a headers dict to the constructor.'
                        )
                    headers[k] = val_1
        return headers

    @staticmethod
    def _check_sample_shapes_and_dtype_consistency(thss_dict):
        key0 = list(thss_dict.keys())[0]
        ths0 = thss_dict[key0]
        for key in thss_dict.keys():
            ths1 = thss_dict[key]
            if ths0._reader.get_trace_size(0) != ths1._reader.get_trace_size(0):
                raise ValueError(f'Inconsistent samples shapes: samples length is {ths0._reader.get_trace_size(0)} for {key0} and samples length '
                                 f'is {ths1._reader.get_trace_size(0)} for {key}')
            if ths0[0].samples[0].dtype != ths1[0].samples[0].dtype:
                raise ValueError(f'Inconsistent samples dtypes: samples dtype is {ths0[0].samples[0].dtype} for {key0} and samples dtype '
                                 f'is {ths1[0].samples[0].dtype} for {key}')

    @staticmethod
    def _check_metadata_shapes_and_dtype_consistency(thss_dict):
        key0 = list(thss_dict.keys())[0]
        ths0 = thss_dict[key0]
        metas0 = list(ths0.metadatas)
        for key in thss_dict.keys():
            ths1 = thss_dict[key]
            for meta in metas0:
                meta0 = ths0[0].metadatas[meta]
                meta1 = ths1[0].metadatas[meta]
                dtype0 = ConcatFormatReader._get_meta_dtype(meta0)
                dtype1 = ConcatFormatReader._get_meta_dtype(meta1)
                if dtype0.kind != dtype1.kind:
                    raise ValueError(f'Inconsistent {meta} dtypes: {meta} dtype is {dtype0} for {key0} and {meta} dtype '
                                     f'is {dtype1} for {key}')
                if dtype0.kind != 'U':
                    if len(meta0) != len(meta1):
                        raise ValueError(f'Inconsistent {meta} shapes: {meta} length is {len(meta0)} for {key0} and {meta} length '
                                         f'is {len(meta1)} for {key}')

    @staticmethod
    def _get_meta_dtype(meta):
        try:
            dtype = meta.dtype
            if dtype.kind == 'O':
                meta = meta.flat[0]
                raise AttributeError
            return meta.dtype
        except AttributeError:
            if not isinstance(meta, str):
                raise TypeError(f'Metadata type not supported: {type(meta)}')
            return _np.dtype(f'<U{len(meta)}')

    def _get_metadatas_infos(self):
        ths = self.ths_list[0]
        metadatas_infos = dict()
        for meta in ths._reader.metadatas_keys:
            meta0 = ths[0].metadatas[meta]
            dtype = self._get_meta_dtype(meta0)
            length = 0 if dtype.kind == 'U' else len(meta0)
            metadatas_infos.update({f'{meta}': dict(length=length, dtype=dtype)})
        return metadatas_infos

    def _convert_traces_indices_to_list_of_raw_traces_indices_per_set(self, traces_indices):
        original_traces_indices = self._sub_traceset_indices[traces_indices]
        cumsizes = _np.cumsum(self._sizes)
        sets = _np.searchsorted(cumsizes, original_traces_indices, side='right')
        indices_per_set = [_np.where(sets == i)[0] for i in range(len(self._sizes))]
        original_traces_indices_per_set = [original_traces_indices[indices] - offset for indices, offset in zip(indices_per_set, _np.concatenate(([0], cumsizes)))] # noqa
        return original_traces_indices_per_set, indices_per_set, len(original_traces_indices)

    def fetch_metadatas(self, key, trace_id):
        if trace_id is None:
            trace_id = slice(len(self))
        is_int_trace_id = False
        if isinstance(trace_id, int):
            trace_id = [trace_id]
            is_int_trace_id = True

        original_traces_indices_per_set, indices_per_set, length = self._convert_traces_indices_to_list_of_raw_traces_indices_per_set(trace_id)
        shape = (length, self._metas_infos[key]['length'])
        shape = (shape[0], ) if shape[1] == 0 else shape
        rez = _np.empty(shape=shape, dtype=self._metas_infos[key]['dtype'])
        for indices, result_indices, ths in zip(original_traces_indices_per_set, indices_per_set, self.ths_list):
            if len(indices):
                data = ths.metadatas[key][indices]
                data_dtype = self._get_meta_dtype(data)
                if data_dtype.itemsize > rez.dtype.itemsize:
                    rez = rez.astype(data_dtype)
                rez[result_indices] = data
        if is_int_trace_id:
            rez = rez[0]
        return rez

    def fetch_samples(self, traces, frame):
        if frame is None:
            frame = slice(None)
        if traces is None:
            traces = slice(len(self))
        if isinstance(traces, int):
            traces = slice(traces, traces + 1)
        if isinstance(frame, int):
            if frame < 0:
                frame = slice(frame, None)
            else:
                frame = slice(frame, frame + 1)

        original_traces_indices_per_set, indices_per_set, length = self._convert_traces_indices_to_list_of_raw_traces_indices_per_set(traces)
        rez = _np.empty(shape=(length, self._trace_size), dtype=self._trace_dtype)
        rez = rez[:, frame]
        for indices, result_indices, ths in zip(original_traces_indices_per_set, indices_per_set, self.ths_list):
            if len(indices):
                data = ths.samples[indices, frame]
                rez[result_indices] = data
        return rez

    def __getitem__(self, key):
        super().__getitem__(key)

        if isinstance(key, int):
            key = [key]

        new_sub_traceset_indices = self._sub_traceset_indices[key]
        new_reader = ConcatFormatReader(check=False, headers=self._headers, **self._ths_dict)
        new_reader._sub_traceset_indices = new_sub_traceset_indices
        new_reader._size = len(new_sub_traceset_indices)

        return new_reader

    @property
    def metadatas_keys(self):
        return self.ths_list[0]._reader.metadatas_keys

    @property
    def headers_keys(self):
        return self._headers.keys()

    def fetch_header(self, key):
        return self._headers[key]

    def get_trace_size(self, trace_id):
        return self._trace_size

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return f'Concat format reader of {self.ths_list} with {len(self)} traces.'


# See https://gitlab.com/eshard/estraces/-/issues/18
@lru_cache(maxsize=8)
def _cached_arange(s):
    return _np.arange(s)



from .context import estraces  # noqa
import pytest
import numpy as np
from .conftest import DATAS


@pytest.fixture(params=['1 trace', 'all traces'])
def samples(request, fmt):
    return estraces.traces.samples.Samples(reader=fmt, trace_id=0 if request.param == '1 trace' else None)


@pytest.fixture
def s1d(fmt):
    return estraces.traces.samples.Samples(reader=fmt, trace_id=0)


@pytest.fixture
def s2d(fmt):
    return estraces.traces.samples.Samples(reader=fmt)


def test_samples_repr_is_equal_to_wrapped_ndarray_repr(samples):
    assert repr(samples) == repr(samples[...])


def test_1d_samples_len_is_equal_to_ndarray_len(s1d):
    assert len(s1d) == len(s1d[...])


def test_2d_samples_len_is_equal_to_number_of_traces(s2d):
    assert 10 == len(s2d)


def test_samples_init_with_improper_reader_type_raises_exception():
    with pytest.raises(TypeError):
        estraces.traces.samples.Samples(reader=lambda s: s)


def test_samples_slicing_with_improper_index_type_raises_error(samples):
    with pytest.raises(IndexError):
        samples["jfkkjdl"]
    with pytest.raises(IndexError):
        samples[..., "klklkg"]
    with pytest.raises(IndexError):
        samples["hjkghkj", ()]
    with pytest.raises(IndexError):
        samples["hjkghkj", "fhfh"]


def test_trace_samples_slicing_with_too_large_index_raises_error(samples):
    with pytest.raises(IndexError):
        samples[4555555]
    with pytest.raises(IndexError):
        samples[[1454544, 98090]]
    with pytest.raises(IndexError):
        samples[[1454544, 98090], ...]
    with pytest.raises(IndexError):
        samples[..., [1454544, 98090]]
    with pytest.raises(IndexError):
        samples[..., range(1454544)]


def test_trace_slicing_with_more_than_available_dimensions_raises_error(s1d, s2d):
    with pytest.raises(IndexError):
        s2d[..., 1, ...]
    with pytest.raises(IndexError):
        s1d[..., 1]


def test_samples_array_attribute(samples):
    assert isinstance(samples.array, np.ndarray)
    assert np.array_equal(samples[:], samples.array)


def test_samples_slicing_with_explicit_indices_on_2_dim(s2d):
    for i in range(10):
        assert len(s2d[i].shape) == 1
        assert np.array_equal(s2d[i, :], DATAS[i, :])
        assert np.array_equal(s2d[i, 10:50], DATAS[i, 10:50])


def test_samples_slicing_with_int_slice_list_on_traces_index(s2d):
    assert np.array_equal(s2d[2], DATAS[2, :])
    assert np.array_equal(s2d[1:8:2], DATAS[1:8:2, :])
    assert np.array_equal(s2d[[1, 2, 3]], DATAS[[1, 2, 3], :])


def test_samples_slicing_with_duped_unordered_list(s1d, s2d):
    assert DATAS[0, np.array([1, 4, 3, 7, 4, 9])].tolist() == s1d[np.array([1, 4, 3, 7, 4, 9])].tolist()
    assert DATAS[[3, 1, 2], :].tolist() == s2d[[3, 1, 2], :].tolist()
    assert DATAS[[3, 1, 2, 1], :].tolist() == s2d[[3, 1, 2, 1], :].tolist()
    assert DATAS[[1, 2, 1], :][:, [7, 8, 7]].tolist() == s2d[[1, 2, 1], [7, 8, 7]].tolist()


def test_samples_slicing_traces_with_upper_bound_out_of_bounds_returns_samples_on_available_traces(s2d):
    assert s2d[5:1000].shape[0] == 5
    assert s2d[11:1000].shape[0] == 0


def test_samples_ranging_traces_with_upper_bound_out_of_bounds_returns_samples_on_available_traces(s2d):
    assert s2d[range(5, 1000)].shape[0] == 5
    assert s2d[range(11, 1000)].shape[0] == 0


def test_samples_slicing_with_int_slice_list_on_samples_index(s2d):
    assert DATAS[:, 1:8:2].tolist() == s2d[:, 1:8:2].tolist()
    assert DATAS[:, 2].tolist() == s2d[:, 2].tolist()
    assert DATAS[:, [1, 2, 3]].tolist() == s2d[:, [1, 2, 3]].tolist()
    assert DATAS[..., [1, 2, 13]].tolist() == s2d[..., [1, 2, 13]].tolist()


def test_samples_iterates_over_traces(s2d):
    for index, trc_samples in enumerate(s2d):
        assert np.array_equal(trc_samples[:], DATAS[index])


def test_2d_samples_with_one_trace_has_appropriate_shape(s2d):
    s2d_bis = s2d[0:1, [1, 2]]
    assert s2d_bis.ndim == 2
    assert s2d_bis.shape[0] == 1

    s2d_bis = s2d[0:1, 1:4]
    assert s2d_bis.ndim == 2
    assert s2d_bis.shape[0] == 1

    s2d_ter = s2d[0, 1:10]
    assert s2d_ter.ndim == 1

    s2d_ter = s2d[0, [1, 2]]
    assert s2d_ter.ndim == 1


def test_trace_samples_iteration(s1d):
    for s in s1d:
        assert s is not None


def test_trace_samples_tolist(s1d):
    assert isinstance(s1d.tolist(), list)


def test_trace_samples_slicing_empty_array_with_out_of_the_window_slice(s1d):
    smp = s1d[5000:4544444]
    assert np.array_equal(np.array([]), smp)


def test_trace_samples_slicing_with_ellipsis_or_none(s1d):
    assert isinstance(s1d[...], np.ndarray)
    assert np.array_equal(s1d[...], s1d[:])


def test_trace_samples_slicing_with_slice(s1d):
    assert isinstance(s1d[10:200:2], np.ndarray)
    assert len(s1d[0:200:2]) == 100


def test_trace_samples_slicing_with_range(s1d):
    assert isinstance(s1d[range(10, 200, 2)], np.ndarray)
    assert len(s1d[range(0, 200, 2)]) == 100


def test_trace_samples_slicing_with_list(s1d):
    assert isinstance(s1d[[10, 20, 40]], np.ndarray)
    assert len(s1d[[10, 20, 40]]) == 3


def test_trace_samples_slicing_with_int(s1d):
    assert s1d[3] == s1d[:][3]
    assert isinstance(s1d[3], np.uint8)
    assert isinstance(s1d[3], type(s1d[:][0]))


def test_trace_samples_slicing_with_out_of_bounds_int(s1d, s2d):
    with pytest.raises(IndexError, match='index 10000000 is out of bounds for axis 0 with size 1000'):
        s1d[10000000]
    with pytest.raises(IndexError, match='index 11 is out of bounds for axis 0 with size 10'):
        s2d[11, :]
    with pytest.raises(IndexError, match='index 10000000000 is out of bounds for axis 1 with size 1000'):
        s2d[:, 10000000000]
    with pytest.raises(IndexError, match='index 10000000000 is out of bounds for axis 1 with size 1000'):
        s2d[0, 10000000000]


def test_samples_slicing_with_slice_of_len_zero(s1d, s2d):
    assert isinstance(s1d[10:10], np.ndarray)
    assert 0 == len(s1d[10:10])

    assert isinstance(s2d[:, 10:10], np.ndarray)
    assert 0 == s2d[:, 10:10].shape[1]

    assert isinstance(s2d[10:10, :], np.ndarray)
    assert 0 == s2d[10:10, :].shape[0]


def test_samples_slicing_with_range_of_len_zero(s1d, s2d):
    assert isinstance(s2d[10:10, :], np.ndarray)
    assert 0 == s2d[10:10, :].shape[0]


def test_samples_frame_slicing_with_range_of_len_zero_raises_index_error(s1d, s2d):
    with pytest.raises(IndexError):
        s1d[range(10, 10)]
    with pytest.raises(IndexError):
        s2d[:, range(10, 10)]


def test_samples_slicing_with_negative_slices(s1d, s2d):
    assert s1d[800:900].tolist() == s1d[-200: -100].tolist()
    assert [] == s1d[-100: -200].tolist()

    assert s2d[:, 800:900].tolist() == s2d[:, -200: -100].tolist()
    assert [[] for i in range(s2d.shape[0])] == s2d[:, -100: -200].tolist()

    assert s2d[8:9, :].tolist() == s2d[-2: -1, :].tolist()
    assert [] == s2d[-1:-2, :].tolist()


def test_samples_slicing_with_negative_range(s1d, s2d):
    assert s1d[800:900].tolist() == s1d[range(-200, -100)].tolist()
    with pytest.raises(IndexError):
        s1d[range(-100, -200)]

    assert s2d[:, 800:900].tolist() == s2d[:, range(-200, -100)].tolist()
    with pytest.raises(IndexError):
        s2d[:, range(-100, -200)]

    assert s2d[8:9, :].tolist() == s2d[range(-2, -1), :].tolist()
    assert [] == s2d[range(-1, -2), :].tolist()


def test_samples_implements_array_interface(s1d, s2d):
    # See issue https://gitlab.eshard.int/esdynamic/estraces/issues/35
    assert s1d.__array_interface__ is not None
    assert s2d.__array_interface__ is not None
    assert s1d.__array_struct__ is not None
    assert s2d.__array_struct__ is not None
    assert isinstance(s1d.__array_interface__, type(s1d[:].__array_interface__)) is not None
    assert isinstance(s2d.__array_interface__, type(s2d[:].__array_interface__)) is not None


def test_samples_is_castable_to_numpy(s1d, s2d):
    assert np.array_equal(
        np.array(s1d),
        s1d[:]
    )
    assert np.array_equal(
        np.array(s2d),
        s2d[:]
    )


def test_samples_instantiate_array_data_in_last_resort(s1d, s2d):
    # See issue https://gitlab.eshard.int/esdynamic/estraces/issues/52
    assert s2d._array is None
    s2d[0]
    assert s2d._array is None
    s2d.array
    assert s2d._array is not None


def test_samples_support_boolean_indexing(samples):
    boolean = np.array([True] + [False] * (len(samples) - 2) + [True], dtype=bool)
    assert len(samples[boolean]) == 2
    assert np.array_equal(samples[0], samples[boolean][0])
    assert np.array_equal(samples[-1], samples[boolean][1])


def test_samples_support_boolean_indexing_2d(s2d):
    boolean = np.array([True] + [False] * (s2d.shape[1] - 2) + [True], dtype=bool)
    assert s2d[:, boolean].shape[1] == 2
    assert np.array_equal(s2d[:, 0], s2d[:, boolean][:, 0])
    assert np.array_equal(s2d[:, -1], s2d[:, boolean][:, 1])


def test_samples_shape_does_not_load_data(s1d, s2d):
    shape = s1d.shape
    assert s1d._array is None
    assert shape == s1d.array.shape

    shape = s2d.shape
    assert s2d._array is None
    assert shape == s2d.array.shape


def test_samples_dtype_does_not_load_data(s1d, s2d):
    dtype = s1d.dtype
    assert s1d._array is None
    assert dtype == s1d.array.dtype

    dtype = s2d.dtype
    assert s2d._array is None
    assert dtype == s2d.array.dtype


def test_samples_size_does_not_load_data(s1d, s2d):
    size = s1d.size
    assert s1d._array is None
    assert size == s1d.array.size

    size = s2d.size
    assert s2d._array is None
    assert size == s2d.array.size



from .context import estraces  # noqa
import pytest
import numpy as np
from .format_test_implementation import Format
from .conftest import DATAS, PLAINS, CIPHERS, KEY, FOO, TIME


@pytest.fixture
def ths(fmt):
    return estraces.traces.trace_header_set.build_trace_header_set(reader=fmt, name="Test ths")


@pytest.fixture
def ths2():
    fmt = Format(metadatas={"other": range(10)})
    return estraces.traces.trace_header_set.build_trace_header_set(reader=fmt, name="Another ths")


def test_trace_header_set_cant_be_instantiate_directly():
    with pytest.raises(TypeError):
        estraces.traces.trace_header_set.TraceHeaderSet(reader={})


def test_read_metadatas_returns_metadatas_type(ths):
    assert isinstance(ths.metadatas, estraces.traces.metadatas.Metadatas)


def test_read_metadatas_with_attribute(ths, fmt):
    assert np.array_equal(ths.metadatas["ciphertext"], ths.ciphertext)
    assert np.array_equal(ths.ciphertext, CIPHERS)
    assert np.array_equal(ths.metadatas["plaintext"], ths.plaintext)
    assert np.array_equal(ths.plaintext, PLAINS)


def test_read_metadatas_attributes_missing_raises_attribute_error(ths, ths2):
    with pytest.raises(AttributeError):
        ths.myproperty
    with pytest.raises(AttributeError):
        ths2.ciphertext


def test_trace_header_set_initialization_with_improper_format():
    with pytest.raises(TypeError):
        estraces.traces.trace_header_set.build_trace_header_set(reader={}, name="")


def test_trace_header_set_initialization(ths):
    assert isinstance(ths, estraces.TraceHeaderSet)


def test_traces_attribute(ths):
    assert isinstance(ths.traces[0], estraces.Trace)
    assert len(ths.traces) == 10


def test_traces_length_when_slicing(ths):
    # See issue https://gitlab.eshard.int/esdynamic/estraces/issues/41
    # with a correct __len__ we are able to iterate over the sliced traces
    assert len(ths.traces[:]) == 10
    assert len(ths.traces[0:]) == 10
    assert len(ths.traces[2:]) == 8
    assert len(ths.traces[:10]) == 10
    assert len(ths.traces[:8]) == 8
    assert len(ths.traces[2:8]) == 6
    assert len(ths.traces[-2:10]) == 2
    assert len(ths.traces[2:-2]) == 6


def test_samples_returns_samples_type(ths):
    assert isinstance(ths.samples, estraces.traces.samples.Samples)


def test_ths_slice_is_a_new_ths(ths):
    new_ths = ths[:]
    assert isinstance(new_ths, estraces.TraceHeaderSet)
    assert id(new_ths) != id(ths)
    for idx, t in enumerate(new_ths.samples):
        assert id(t) != id(ths.samples[idx])
        assert np.array_equal(t[:], ths.samples[idx])


def test_ths_slicing_with_out_of_range_int_raises_index_error(ths):
    with pytest.raises(IndexError):
        ths[100000]


def test_ths_slicing_raises_error_with_improper_types(ths):
    with pytest.raises(IndexError):
        ths[1, 44]
    with pytest.raises(IndexError):
        ths["kdjfkldjflk"]
    with pytest.raises(IndexError):
        ths[range(1, 10)]
    with pytest.raises(IndexError):
        ths[np.array([[1, 2], [3, 4]])]
    with pytest.raises(IndexError):
        ths[np.array(["x", "tutu"])]
    with pytest.raises(IndexError):
        ths[np.array([1.4, 2.2])]
    with pytest.raises(IndexError):
        ths[None]
    with pytest.raises(IndexError):
        ths[np.float32(1.6)]


def test_ths_slicing_int_or_numpy_int(ths):
    expected = DATAS[1]
    for t in (1, np.int8(1), np.int32(1), np.int64(1)):
        assert np.array_equal(ths[t].samples[:], expected)


def test_ths_iteration_on_traces(ths):
    for t in ths:
        assert isinstance(t, estraces.Trace)


def test_ths_slicing_with_int_returns_a_trace(ths):
    assert isinstance(ths[1], estraces.Trace)
    assert np.array_equal(ths[1].samples[:], DATAS[1])


def test_ths_slicing_with_out_of_bounds_value_raise_exception(ths):
    new_ths = ths[[0, 2, 4, 6, 8, 9]]

    with pytest.raises(IndexError):
        new_ths[10]

    with pytest.raises(IndexError):
        new_ths[[1, 11]]

    with pytest.raises(IndexError):
        new_ths[np.array([1, 11])]


def test_ths_slicing_with_exotic_slices_returns_ths_of_max_size(ths):
    assert 10 == len(ths[:1000])
    assert 8 == len(ths[-11:8])
    assert 10 == len(ths[-1000:])
    assert 5 == len(ths[-1000:-5])
    assert 0 == len(ths[:-12])
    assert 3 == len(ths[-5: -2])
    assert 8 == len(ths[-12: -2])
    assert 0 == len(ths[-2: -12])
    assert 0 == len(ths[-2: 1])
    assert 1 == len(ths[-2: 9])
    assert 2 == len(ths[-2: 10])
    assert 10 == len(ths[-1000:1000])
    assert 0 == len(ths[1000:-1000])
    assert 0 == len(ths[1000:-5])
    assert 0 == len(ths[1000:])
    assert 0 == len(ths[1000:2])
    assert 0 == len(ths[1000:2000])


def test_ths_slicing_with_slice_returns_subths(ths):
    new_ths = ths[0:2:2]
    assert len(new_ths) == 1
    new_ths = ths[0:7:2]
    assert len(new_ths) == 4
    assert new_ths.plaintext.tolist() == ths.plaintext[0:7:2].tolist()

    new_ths = ths[3:]
    assert len(new_ths) == 7
    new_ths = ths[:8]
    assert len(new_ths) == 8


def test_succesive_ths_slicings_keeps_samples_metas_consistent(ths):
    sub_t_1 = ths[0:4]
    sub_t_2 = ths[4:8]
    assert len(sub_t_1) == 4
    assert len(sub_t_2) == 4
    assert sub_t_1.samples.tolist() == ths.samples[0:4].tolist()
    assert sub_t_2.samples.tolist() == ths.samples[4:8].tolist()
    assert sub_t_1.plaintext.tolist() == ths.plaintext[0:4].tolist()
    assert sub_t_2.plaintext.tolist() == ths.plaintext[4:8].tolist()
    assert sub_t_1.ciphertext.tolist() == ths.ciphertext[0:4].tolist()
    assert sub_t_2.ciphertext.tolist() == ths.ciphertext[4:8].tolist()
    assert sub_t_1.headers == sub_t_2.headers
    assert ths.headers == sub_t_1.headers

    sub_sub_t_1 = sub_t_1[0:2]
    sub_sub_t_2 = sub_t_2[0:2]
    assert len(sub_sub_t_1) == 2
    assert len(sub_sub_t_2) == 2
    assert sub_sub_t_1.samples.tolist() == ths.samples[0:2].tolist()
    assert sub_sub_t_2.samples.tolist() == ths.samples[4:6].tolist()
    assert sub_sub_t_1.plaintext.tolist() == ths.plaintext[0:2].tolist()
    assert sub_sub_t_2.plaintext.tolist() == ths.plaintext[4:6].tolist()
    assert sub_sub_t_1.plaintext.tolist() == ths.plaintext[0:2].tolist()
    assert sub_sub_t_2.plaintext.tolist() == ths.plaintext[4:6].tolist()
    assert sub_sub_t_1.ciphertext.tolist() == ths.ciphertext[0:2].tolist()
    assert sub_sub_t_2.ciphertext.tolist() == ths.ciphertext[4:6].tolist()
    assert sub_sub_t_1.headers == sub_sub_t_2.headers
    assert ths.headers == sub_sub_t_1.headers

    sub_sub_t_12 = sub_t_1[2:4]
    assert len(sub_sub_t_12) == 2
    assert sub_sub_t_12.samples.tolist() == ths.samples[2:4].tolist()
    assert sub_sub_t_12.plaintext.tolist() == ths.plaintext[2:4].tolist()
    assert sub_sub_t_12.ciphertext.tolist() == ths.ciphertext[2:4].tolist()
    assert sub_sub_t_12.headers == ths.headers

    sub_sub_t_22 = sub_t_2[2:4]
    assert len(sub_sub_t_22) == 2
    assert sub_sub_t_22.samples.tolist() == ths.samples[6:8].tolist()
    assert sub_sub_t_22.plaintext.tolist() == ths.plaintext[6:8].tolist()
    assert sub_sub_t_22.ciphertext.tolist() == ths.ciphertext[6:8].tolist()
    assert sub_sub_t_22.headers == ths.headers

    sub_sub_sub_t_1 = sub_sub_t_1[0:1]
    assert len(sub_sub_sub_t_1) == 1
    assert sub_sub_sub_t_1.samples.tolist() == ths.samples[0:1].tolist()
    assert sub_sub_sub_t_1.plaintext.tolist() == ths.plaintext[0:1].tolist()
    assert sub_sub_sub_t_1.ciphertext.tolist() == ths.ciphertext[0:1].tolist()
    assert sub_sub_sub_t_1.headers == ths.headers


def test_ths_slicing_with_list_returns_subths(ths):
    new_ths = ths[[1, 2, 4]]
    assert len(new_ths) == 3


def test_ths_slicing_with_nparray_returns_subths(ths):
    new_ths = ths[np.array([1, 2, 4])]
    assert len(new_ths) == 3


def test_ths_slicing_with_undeduped_unordered_list(ths):
    new_ths = ths[[3, 2, 1]]
    assert 3 == len(new_ths)
    assert ths[3].samples[:].tolist() == new_ths[0].samples[:].tolist()
    assert ths[1].samples[:].tolist() == new_ths[-1].samples[:].tolist()

    new_ths = ths[[3, 2, 1, 2]]
    assert 4 == len(new_ths)
    assert ths[3].samples[:].tolist() == new_ths[0].samples[:].tolist()
    assert ths[2].samples[:].tolist() == new_ths[-1].samples[:].tolist()
    assert ths[2].samples[:].tolist() == new_ths[1].samples[:].tolist()
    assert ths[1].samples[:].tolist() == new_ths[2].samples[:].tolist()


def test_ths_slicing_with_zero_len_slice_returns_empty_ths(ths):
    assert isinstance(ths[0:0], estraces.TraceHeaderSet)
    assert 0 == len(ths[0:0])


def test_ths_filter():
    ths = estraces.traces.trace_header_set.build_trace_header_set(
        name="test ths",
        reader=Format(datas=np.vstack([np.array(range(i, i + 100)) for i in range(100)])),
    )
    filtered_ths = ths.filter(lambda t: t.samples[0] % 2 == 0)
    assert len(filtered_ths) == 50
    assert id(ths) != filtered_ths
    for trc in filtered_ths:
        assert trc.samples[0] % 2 == 0


def test_split(ths):
    split = ths.split(part_size=2)
    assert len(split) == 5

    for t in split:
        assert len(t) == 2

    split = ths.split(part_size=3)
    assert len(split) == 4

    for idx in range(len(split) - 1):
        assert len(split[idx]) == 3
        assert len(split[idx]) == split[idx].samples[:].shape[0]
    assert len(split[-1]) == 1
    assert len(split[-1]) == split[-1].samples[:].shape[0]

    for t in split:
        assert isinstance(t, estraces.TraceHeaderSet)


def test_split_raises_error_with_improper_type(ths):
    with pytest.raises(TypeError):
        ths.split(part_size="eeze")
        ths.split(part_size=-4)

    split = ths.split(part_size=2)
    with pytest.raises(TypeError):
        split["jj"]
        split[-15]
        split[1:2]


def test_str_and_repr(ths):
    assert repr(ths) == str(ths)
    assert repr(ths[0]) == str(ths[0])


def test_attribute_headers_returns_headers_type(ths):
    assert isinstance(ths.headers, estraces.traces.headers.Headers)


def test_attribute_headers(ths):
    assert np.array_equal(ths.headers['key'], KEY)
    assert ths.headers['time'] == TIME
    assert ths.headers['foo'] == FOO


def test_add_metadata_to_ths(ths):
    datas = np.random.randint(0, 255, (len(ths), 1), dtype='uint8')
    ths.metadatas['new_meta'] = datas
    assert np.array_equal(ths.metadatas['new_meta'], datas)
    assert np.array_equal(ths.new_meta, datas)


def test_init_metadatas_and_slice_ths_works_as_expected(ths):
    ths.metadatas
    assert np.array_equal(ths[:1].metadatas['plaintext'], PLAINS[:1])


def test_add_metadata_to_ths_through_attribute(ths):
    datas = np.random.randint(0, 255, (len(ths), 1), dtype='uint8')
    ths.new_meta = datas
    assert np.array_equal(ths.metadatas['new_meta'], datas)
    assert np.array_equal(ths.new_meta, datas)

    sub_ths = ths[0:2]
    assert np.array_equal(sub_ths.metadatas['new_meta'], datas[:2])
    assert np.array_equal(sub_ths.new_meta, datas[:2])

    trace = ths[8]
    assert np.array_equal(trace.metadatas['new_meta'], datas[8])
    assert np.array_equal(trace.new_meta, datas[8])


def test_add_existing_metadata_or_attribute_raises_exception(ths):
    with pytest.raises(AttributeError):
        ths.plaintext = 'foo'
    with pytest.raises(AttributeError):
        ths.samples = 'foo'
    ths.__doc__ = 'foo'
    assert ths.__doc__ == 'foo'
    with pytest.raises(KeyError):
        ths.metadatas['__doc__']



import numpy as _np

from estraces.traces.abstract_reader import AbstractReader as _AbstractReader
from estraces.traces.trace_header_set import build_trace_header_set as _build_trace_header_set


def read_ths_from_random(nb_traces: int,
                         trace_length: int,
                         interval=(-128, 128),
                         dtype='int8',
                         seed=None,
                         headers={},
                         **kwargs):
    """Allow user to a TraceHeaderFile for random trace reader.

    Attributes:
        nb_trace (int): Number of trace to randomly generate
        trace_length (int): Length of trace to generate
        interval : default value are (-128, 128)
        dtype (str): default value is 'int8'
        seed : seed to use for `_np.random.seed(seed)`
        **kwargs (ndarray): metadata as ndarray, for example, plaintext=(16, 'uint8')

    """
    return _build_trace_header_set(reader=RandomReader(nb_traces=nb_traces,
                                                       trace_length=trace_length,
                                                       interval=interval,
                                                       dtype=dtype,
                                                       seed=seed,
                                                       headers=headers,
                                                       **kwargs),
                                   name='Random reader')


"""RandomReader generate random trace with specific length.
"""


class RandomReader(_AbstractReader):
    """Allow user to generate randomized traces with specific length.

    Attributes:
        nb_trace (int): Number of trace to randomly generate
        trace_length (int): Length of trace to generate
        interval : default value are (-128, 128)
        dtype (str): default value is 'int8'
        seed : seed to use for `_np.random.seed(seed)`
        **kwargs (ndarray): metadata as ndarray, for example plaintext=(16, 'uint8')

    """

    def __init__(self,
                 nb_traces: int,
                 trace_length: int,
                 interval=(-128, 128),
                 dtype='int8',
                 seed=None,
                 headers={},
                 **kwargs
                 ):
        self.nb_traces = nb_traces
        self.trace_length = trace_length
        self.interval = interval
        self.dtype = _np.dtype(dtype)
        self._check_dtype_and_interval()
        _np.random.seed(seed)
        self._headers = headers
        self._metadatas = kwargs

    def _generate_data(self, shape, dtype):
        if dtype == _np.dtype('float64'):
            x, y = _np.random.rand(shape[0]) * 2 - 1, _np.random.rand(shape[1]) * 2 - 1
        else:
            x, y = _np.random.rand(shape[0]).astype('float32') * 2 - 1, _np.random.rand(shape[1]).astype('float32') * 2 - 1
        data = _np.outer(x, y)
        a, b = self.interval
        data = data * ((b - a) / 2) + (b - (b - a) / 2)
        return data.astype(dtype)

    def _check_dtype_and_interval(self):
        if self.interval[0] >= self.interval[1]:
            raise ValueError(f'Minimum must be lower than Maximum, but {self.interval[0]} >= {self.interval[1]}.')
        try:
            infos = _np.iinfo(self.dtype)
        except ValueError:
            infos = _np.finfo(self.dtype)
        if self.interval[0] < infos.min:
            raise ValueError(f'Given minimum value ({self.interval[0]}) is lower than the minimum of the type {self.dtype}: {infos.min}.')
        if self.interval[1] - 1 > infos.max:
            raise ValueError(f'Given maximum value ({self.interval[1]}) is greater than the maximum of the type {self.dtype}: {infos.max}.')

    def _slice_length(self, frame, reference_length):
        start = frame.start if frame.start is not None else 0
        step = frame.step if frame.step is not None else 1
        stop = frame.stop
        if stop is None or stop > reference_length:
            stop = reference_length
        return abs(stop - start) // abs(step)

    def _key_length(self, key, reference_length):
        if isinstance(key, (slice, range)):
            return self._slice_length(key, reference_length)
        if key is Ellipsis:
            return reference_length
        return len(key)

    def fetch_samples(self, traces=slice(None), frame=slice(None)):
        nb_traces = self._key_length(traces, self.nb_traces)
        trace_length = self._key_length(frame, self.trace_length)
        return self._generate_data((nb_traces, trace_length), self.dtype)

    def fetch_metadatas(self, key, trace_id=None):
        length, dtype = self._metadatas[key]
        nb_data = self.nb_traces if trace_id is None else 1
        data = self._generate_data((nb_data, length), dtype)
        if trace_id is not None:
            return data[:, 0]
        return data

    @property
    def metadatas_keys(self):
        return self._metadatas.keys()

    def __len__(self):
        return self.nb_traces

    def __getitem__(self, key):
        nb_traces = self._key_length(key, self.nb_traces)
        return RandomReader(nb_traces, self.trace_length, self.interval, self.dtype, None, self._headers, **self._metadatas)

    def get_trace_size(self, trace_id=None):
        return self.trace_length

    def fetch_header(self, key):
        return self._headers[key]

    @property
    def headers_keys(self):
        return self._headers.keys()

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f'Random reader with {len(self)} traces of {self.trace_length} length. Interval {self.interval}. Metadatas: {list(self.metadatas_keys)}'



# -*- coding: utf-8 -*-
import os as _os
import binascii as _binascii
import re as _re
import numpy as _np


class DirectValue(object):
    """Object used to handle user defined values for metadata.

    Attributes:
        value       : the direct value to set as a metadata
        unhexlify   : get the binary data out of the hexadecimal representation of value (default: True)

    When unhexlify is True (default), the expected inputs are either a hexadecimal string without the '0x' prefix or
    an integer list in range(0, 256). When unhexlify is False, it can be any regular string.

    Returns:
        a DirectValue() object

    Usage example:
        d0 = DirectValue("00AA11BB")
        d1 = DirectValue([0, 170, 17, 187])
        d2 = DirectValue("John Dow", unhexlify=False)

    """

    def __init__(self, value=None, unhexlify=True):
        self.unhexlify = unhexlify
        self.value = self._check_value(value)

    def _check_value(self, input):
        """Check that the user input is valid."""
        if isinstance(input, str):
            # We expect a valid hexString
            if self.unhexlify:
                # A valid hexString must have an even size
                if len(input) % 2 != 0:
                    raise ValueError("Odd-length string.")
                # and only the following subset of characters
                elif not _re.fullmatch(r"[0-9a-fA-F]*", input):
                    raise ValueError("Invalid character in input string.")
            # Any regular string is fine otherwise
        elif isinstance(input, list):
            try:
                bytes(input)
            except TypeError as t_err:
                raise TypeError(t_err)
        else:
            raise ValueError(f"Invalid value format {type(input)}. HexString without the '0x' prefix or a list of integer in range(0, 256) is expected.")

        return input

    def get_text(self, _):
        """Get the user input to be used as Direct Value.

        Returns:
            byte array

        """
        # User input is a string
        if isinstance(self.value, str):
            if self.unhexlify:
                return _binascii.unhexlify(self.value)
            else:
                return self.value
        # User input is an integer list
        elif isinstance(self.value, list):
            if self.unhexlify:
                return bytes(self.value)
            else:
                return str(self.value)


class HeaderExtractor(object):
    """Object to retrieve metadata from binary files.

    The user must specify a start reading offset, and either the end offset
    or the number of bytes to read.

    Attributes:
        start   : start reading offset
        end     : end reading offset
        count   : the number of byte to read

    Usage:
        plain = HeaderExtractor(start=0, end=15)
        cipher = HeaderExtractor(start=16, count=16)

    """

    def __init__(self, start=None, end=None, count=None, unhexlify=True):
        self._check_and_set_boundaries(start, end, count)
        self.unhexlify = unhexlify

    def _check_and_set_boundaries(self, start, end, count):
        if start is not None:
            if end is None and count is None:
                raise ValueError("No 'end' offset or 'count' specified.")
            elif end is not None and count is not None:
                raise ValueError("Use either 'end' or 'count' option, not both.")
        else:
            raise ValueError("No 'start' offset specified.")

        if not isinstance(start, int) or start < 0:
            raise ValueError('start offset must be positive integer.')

        self.start = start

        if end is not None:
            if not isinstance(end, int) or end < 0:
                raise ValueError('end offset must be positive integer.')
            self.end = end
            self.count = self.end - self.start

        if count is not None:
            if not isinstance(count, int) or count < 0:
                raise ValueError('count must be positive integer.')
            self.count = count
            self.end = self.start + count

        if self.end <= self.start:
            raise ValueError('The end offset must be greater than the start offset.')

    def get_text(self, filename):
        """Get the text from a binary hexadecimal string contained in a file header."""
        file_size = _os.stat(filename).st_size
        if self.end > file_size:
            raise ValueError(f'{self.end} is greater than file size ({file_size} bytes)')
        file_pointer = _np.memmap(filename, dtype=_np.uint8, mode='r')
        data = _np.copy(file_pointer[self.start:self.end])
        del file_pointer

        if not self.unhexlify:
            return str("".join(hex(i)[2:] for i in data))
        else:
            return data.tobytes()

    def get_text_stacked(self, header):
        """Get the text from a binary header contained in a stacked file."""
        if self.end > len(header):
            raise ValueError(f'{self.end} is greater than header size ({len(header)} bytes).')
        data = _np.copy(header[self.start:self.end])

        if not self.unhexlify:
            return str("".join(hex(i)[2:] for i in data))
        else:
            return data.tobytes()


class PatternExtractor(object):
    """Object using regular expressions to parse strings.

    Attributes:
        pattern     : a string specifying the regular expression.
        replace     : the replacement pattern using regex groups.
        unhexlify   : returns the binary data represented by the hexadecimal string value.

    """

    def __init__(self, pattern=None, replace=None, num=0, unhexlify=True):
        self.pattern = _re.compile(pattern)
        self.replace = replace
        self.num = num
        self.unhexlify = unhexlify

    def get_pattern(self, string):
        """Return the matching PatternExtractor.

        The pattern that matches the regex self.pattern in the input string will be returned.
        According to self.unhexlify, it will return a string or binary data.
        """
        string = string.rstrip()
        if self.replace:
            string = self.pattern.sub(self.replace, string)
        elif self.num is not None:
            try:
                string = self.pattern.findall(string)[self.num]
            except IndexError:
                raise ValueError(f"Pattern '{self.pattern.pattern}' not found in '{string}'.")
        if self.unhexlify:
            string = _binascii.unhexlify(string)
        return string

    def get_text(self, filename):
        return self.get_pattern(self.get_value(filename))

    def get_value(self, filename):
        # FIXME: What is the purpose of this function ?
        return filename


class FilePatternExtractor(PatternExtractor):
    """Object iterating PatternExtractor to parse files."""

    def __init__(self, filename, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filename = self._check_filename(filename)
        self._iter = self.text_generator(filename)
        self._fixed_value = None

    def _check_filename(self, filename):
        if _os.path.isfile(filename):
            return filename
        else:
            raise FileNotFoundError(f"'{filename}' not found")

    def text_generator(self, filename):
        with open(filename, 'r') as f:
            lines = f.readlines()
            if len(lines) > 1:
                for m in lines:
                    yield m
            else:
                self._fixed_value = lines[0]
                yield lines[0]

    def get_value(self, filename):
        if self._fixed_value:
            return self._fixed_value
        else:
            return next(self._iter)



import numpy as _np
from estraces import ETSWriter as _ETSWriter


_SAMPLES_KEY = 'samples'


class _Dataset():
    def __init__(self, length):
        self.length = length
        self.data = None
        self.current_index = None

    def write(self, data):
        """Write data at current dataset index.

        Args:
            data (ndarray): Must be 1D.

        """
        if not isinstance(data, _np.ndarray):
            data = _np.array(data)
        if data.ndim > 1:
            raise ValueError(f'Data must be at most 1D, but {data.ndim}D found')
        if self.data is None:
            self._init(data)

        if self.current_index >= self.length:
            raise ValueError('Dataset is full.')

        if self.data.shape[1] != 1:
            self.data[self.current_index, :len(data)] = data
        else:
            self.data[self.current_index] = data
        self.current_index += 1

    def _init(self, data):
        if data.ndim:
            self.data = _np.zeros_like(data, shape=(self.length, len(data)))
        else:
            self.data = _np.zeros_like(data, shape=(self.length, 1))
        self.current_index = 0

    def _reset(self):
        self.data = None
        self.current_index = None

    def get_data(self):
        """Return data[:current_index + 1] and reset."""
        if self.current_index is None or self.current_index == 0:
            to_return = None
        else:
            to_return = self.data[:self.current_index]
        self._reset()
        return to_return


class BufferedETSWriter():
    """Extend ETSWriter capabilities by adding a bufferization feature."""

    def __init__(self, filename, overwrite=False, buffer_length=None):
        """Create a Buffered ETS file writer instance.

        Args:
            filename (str): Path and filename to write the ETS.
            overwrite (bool, default=False): If True, any existing file with filename will be erased before writing data.
            buffer_length (int): Length of the buffer, in number of traces. If None, the buffer size is computed accordingly
                to the trace size to be equal to 100MB.

        Once the buffer is full, all the data are write back into the ETS file and the buffer is reseted.

        Note that this class provide only "atomic" operations, where samples and metadata are write together, in order to garanty consistency
             between samples and metadata.

        """
        self.ets_writer = _ETSWriter(filename, overwrite=overwrite)
        self.buffer_length = buffer_length
        self.datasets = {}
        self.cnt = 0

    def write_headers(self, headers):
        self.ets_writer.write_headers(headers)

    def write_trace_object_and_points(self, trace_object, points, index=None):
        """Add provided trace samples and metadata to BufferedETS.

        Args:
            trace_object (`Trace`): A `Trace` instance.
            points (numpy.ndarray): Samples to write.
            index (int, default=None): Dummy argument, not taken into account, for compatibility with simple `ETSWriter` format.

        """
        if self.buffer_length is None:
            self.buffer_length = int(100e6 / _np.array(points).nbytes)
        self._write_to_dataset(_SAMPLES_KEY, points)
        for tag, value in trace_object.metadatas.items():
            self._write_to_dataset(tag, value)
        self.cnt += 1
        if self.cnt == self.buffer_length:
            self.flush()

    def flush(self):
        """Write the temporary data into the ETS file."""
        for name1, dataset1 in self.datasets.items():
            for name2, dataset2 in self.datasets.items():
                if dataset1.current_index != dataset2.current_index:
                    raise ValueError('All Datasets must have the same length, '
                                     f'but len({name1})={dataset1.current_index} and len({name2})={dataset2.current_index}')
        for name, dataset in self.datasets.items():
            data = dataset.get_data()
            if data is None:
                return
            if name == _SAMPLES_KEY:
                self.ets_writer.write_samples(data)
            else:
                if data.dtype.kind in {'U', 'S'}:
                    for d in data.squeeze():
                        self.ets_writer.write_metadata(name, str(d))
                else:
                    self.ets_writer.write_metadata(name, data)
        self.cnt = 0

    def _write_to_dataset(self, dataset_name, data):
        if dataset_name not in self.datasets:
            self.datasets[dataset_name] = _Dataset(self.buffer_length)
        self.datasets[dataset_name].write(data)

    def close(self):
        self.flush()
        self.ets_writer.close()

    def get_reader(self):
        self.flush()
        return self.ets_writer.get_reader()



from .context import estraces # noqa
import pytest
import numpy as np
from estraces.formats import ets_format
import cloudpickle

path = 'tests/samples/test.ets'


@pytest.fixture
def ets():
    return ets_format.read_ths_from_ets_file(filename=path)


def compare_ets(ets1, ets2, attribute_names=['plaintext', 'foo_bar', 'ciphertext']):
    assert np.array_equal(ets1.samples[:], ets2.samples[:])
    for attribute_name in attribute_names:
        assert np.array_equal(getattr(ets1, attribute_name), getattr(ets2, attribute_name)), f'{attribute_name} not equal'


def test_serialization_full(ets):
    dumped = cloudpickle.dumps(ets)
    loaded_ets = cloudpickle.loads(dumped)
    compare_ets(ets, loaded_ets)


def test_serialization_sub_ths_part_1(ets):
    dumped = cloudpickle.dumps(ets[10:20])
    loaded_ets = cloudpickle.loads(dumped)
    compare_ets(ets[10:20], loaded_ets)


def test_serialization_sub_ths_part_2(ets):
    ets_part = ets[50:80:2]
    dumped = cloudpickle.dumps(ets_part)
    loaded_ets = cloudpickle.loads(dumped)
    compare_ets(ets_part, loaded_ets)


def test_serialize_headers():
    ets = ets_format.read_ths_from_ets_file('tests/samples/test_with_headers.ets')
    headers_dict = {'foo': 'bar', 'bar': 42, 'list': [1, 2, 3], 'ndarray': np.arange(5, dtype='int8')}
    assert ets.headers == headers_dict


def test_cached_metadata_not_serialized(ets):
    assert ets._metadatas is None
    dumped1 = cloudpickle.dumps(ets)
    assert ets._metadatas is None
    _ = ets.plaintext[:]
    assert ets._metadatas is not None
    dumped2 = cloudpickle.dumps(ets)
    assert len(dumped1) == len(dumped2)
    loaded_ets = cloudpickle.loads(dumped2)
    assert loaded_ets._metadatas is None



from .context import estraces # noqa
import pytest
import sqlite3
import numpy as np
from estraces.formats import sqlite_format
from .conftest import PLAINS, CIPHERS, DATAS, _HEADERS
import os
import timeit


test_path = 'tests/samples/test.db'


@pytest.fixture
def filename():
    try:
        conn = sqlite3.connect(test_path)
        create_samples = '''
        CREATE TABLE traces (
            trace_id INTEGER PRIMARY KEY NOT NULL,
            samples BLOB,
            plaintext BLOB,
            ciphertext BLOB
        )
        '''
        cur = conn.cursor()
        cur.execute(create_samples)
        create_headers = '''
        CREATE TABLE headers (
            id INTEGER PRIMARY KEY NOT NULL,
            tag VARCHAR,
            dtype VARCHAR,
            value BLOB
        )
        '''
        cur = conn.cursor()
        cur.execute(create_headers)

        insert = '''
        INSERT INTO traces(trace_id, samples, plaintext, ciphertext) VALUES (?, ?, ?, ?)
        '''
        for i in range(len(DATAS)):
            cur = conn.cursor()
            cur.execute(insert, (i, DATAS[i], PLAINS[i], CIPHERS[i]))

        insert = 'INSERT INTO headers(id, tag, value, dtype) VALUES (?, ?, ?, ?)'
        i = 1
        for k, v in _HEADERS.items():
            cur = conn.cursor()
            cur.execute(insert, (i, k, v, str(v.dtype) if isinstance(v, np.ndarray) else 'str'))
            i += 1
        conn.commit()
        yield test_path
    finally:
        if conn:
            conn.close()
        os.remove(test_path)


def test_ths_without_headers(filename):
    ths = estraces.read_ths_from_sqlite(
        filename,
        samples_config=sqlite_format.SamplesConfig(),
        metadata_config=sqlite_format.MetadataConfig(metadata_defs={'plaintext': 'uint8'})
    )
    assert isinstance(ths, estraces.TraceHeaderSet)
    assert len(ths) == 10
    assert [] == list(ths.headers.keys())
    assert not ths._reader.fetch_header('tag')


def test_ths_without_metadata(filename):
    ths = estraces.read_ths_from_sqlite(
        filename,
        samples_config=sqlite_format.SamplesConfig()
    )
    assert isinstance(ths, estraces.TraceHeaderSet)
    assert len(ths) == 10
    assert [] == list(ths.metadatas.keys())
    assert not ths._reader.fetch_metadatas('plain')


def test_raises_exception_if_file_does_not_exist():
    with pytest.raises(ValueError):
        estraces.read_ths_from_sqlite('ouch', samples_config=sqlite_format.SamplesConfig())


def test_raises_exception_if_file_is_not_a_sqlite_db():
    with pytest.raises(ValueError):
        estraces.read_ths_from_sqlite('tests/samples/test.ets', samples_config=sqlite_format.SamplesConfig())


def test_raises_exception_fetching_non_existing_header(filename):
    samples_config = sqlite_format.SamplesConfig(table_name='traces', pk_column_name='trace_id', dtype='uint8', samples_column_name='samples')
    metadata_config = sqlite_format.MetadataConfig(
        table_name='traces',
        pk_column_name='trace_id',
        metadata_defs={
            'plaintext': 'uint8',
            'ciphertext': 'uint8',
        }
    )
    headers_config = sqlite_format.HeadersConfig(table_name='headers', tag_column_name='tag', value_column_name='value', dtype_column_name='dtype')

    ths = estraces.read_ths_from_sqlite(
        filename,
        samples_config=samples_config,
        metadata_config=metadata_config,
        headers_config=headers_config
    )
    with pytest.raises(KeyError):
        ths.headers['woop']


def test_raises_exception_with_wrong_header_config(filename):
    samples_config = sqlite_format.SamplesConfig(table_name='traces', pk_column_name='trace_id', dtype='uint8', samples_column_name='samples')
    metadata_config = sqlite_format.MetadataConfig()
    headers_config = sqlite_format.HeadersConfig(table_name='headers', tag_column_name='tags', value_column_name='value', dtype_column_name='dtype')

    ths = estraces.read_ths_from_sqlite(
        filename,
        samples_config=samples_config,
        metadata_config=metadata_config,
        headers_config=headers_config
    )
    with pytest.raises(sqlite_format.SQLiteError):
        ths.headers['key']

    with pytest.raises(sqlite_format.SQLiteError):
        ths._reader.fetch_header('key')


def test_raises_exception_fetching_non_existing_metadata(filename):
    samples_config = sqlite_format.SamplesConfig(table_name='traces', pk_column_name='trace_id', dtype='uint8', samples_column_name='samples')
    metadata_config = sqlite_format.MetadataConfig(
        table_name='traces',
        pk_column_name='trace_id',
        metadata_defs={
            'plaintext': 'uint8',
            'ciphertext': 'uint8',
        }
    )
    headers_config = sqlite_format.HeadersConfig()

    ths = estraces.read_ths_from_sqlite(
        filename,
        samples_config=samples_config,
        metadata_config=metadata_config,
        headers_config=headers_config
    )
    with pytest.raises(KeyError):
        ths.metadatas['woop']


def test_raises_exception_configuring_non_existing_metadata(filename):
    samples_config = sqlite_format.SamplesConfig(table_name='traces', pk_column_name='trace_id', dtype='uint8', samples_column_name='samples')
    metadata_config = sqlite_format.MetadataConfig(
        table_name='traces',
        pk_column_name='trace_id',
        metadata_defs={
            'plaintext': 'uint8',
            'ciphertext': 'uint8',
            'woop': 'uint8'
        }
    )
    headers_config = sqlite_format.HeadersConfig()

    ths = estraces.read_ths_from_sqlite(
        filename,
        samples_config=samples_config,
        metadata_config=metadata_config,
        headers_config=headers_config
    )
    with pytest.raises(sqlite_format.SQLiteError):
        ths[0].metadatas['woop']
    with pytest.raises(sqlite_format.SQLiteError):
        ths.metadatas['woop']


def test_raises_exception_configuring_non_existing_samples_col(filename):
    samples_config = sqlite_format.SamplesConfig(
        table_name='traces',
        pk_column_name='trace_id',
        dtype='uint8',
        samples_column_name='sampless'
    )
    metadata_config = sqlite_format.MetadataConfig()
    headers_config = sqlite_format.HeadersConfig()

    ths = estraces.read_ths_from_sqlite(
        filename,
        samples_config=samples_config,
        metadata_config=metadata_config,
        headers_config=headers_config
    )
    with pytest.raises(sqlite_format.SQLiteError):
        ths[0].samples[:]
    with pytest.raises(sqlite_format.SQLiteError):
        ths.samples[:]
    with pytest.raises(sqlite_format.SQLiteError):
        ths._reader.get_trace_size(0)

    samples_config = sqlite_format.SamplesConfig(
        table_name='tracess',
        pk_column_name='trace_id',
        dtype='uint8',
        samples_column_name='samples'
    )

    ths = estraces.read_ths_from_sqlite(
        filename,
        samples_config=samples_config,
        metadata_config=metadata_config,
        headers_config=headers_config
    )
    with pytest.raises(sqlite_format.SQLiteError):
        ths[0].samples[:]
    with pytest.raises(sqlite_format.SQLiteError):
        ths.samples[:]
    with pytest.raises(sqlite_format.SQLiteError):
        ths._reader.get_trace_size(0)


@pytest.fixture
def lots_of_traces():
    try:
        n_traces = 10000
        l_traces = 200
        conn = sqlite3.connect('lots_of_traces.db')
        create_samples = '''
        CREATE TABLE traces (
            trace_id INTEGER PRIMARY KEY NOT NULL,
            samples BLOB,
            plaintext BLOB
        )
        '''
        cur = conn.cursor()
        cur.execute(create_samples)
        create_headers = '''
        CREATE TABLE headers (
            id INTEGER PRIMARY KEY NOT NULL,
            tag VARCHAR,
            dtype VARCHAR,
            value BLOB
        )
        '''
        cur = conn.cursor()
        cur.execute(create_headers)

        insert = '''
        INSERT INTO traces(trace_id, samples, plaintext) VALUES (?, ?, ?)
        '''
        datas = np.random.randint(0, 255, (n_traces, l_traces), dtype='uint8')
        plains = np.random.randint(0, 255, (n_traces, 16), dtype='uint8')

        for i in range(len(datas)):
            cur = conn.cursor()
            cur.execute(insert, (i, datas[i], plains[i]))
        conn.commit()
        yield 'lots_of_traces.db'
    finally:
        if conn:
            conn.close()
        os.remove('lots_of_traces.db')


def test_samples_shape_is_not_too_slow(lots_of_traces):

    ths = estraces.read_ths_from_sqlite(
        lots_of_traces,
        samples_config=sqlite_format.SamplesConfig(),
        metadata_config=sqlite_format.MetadataConfig(metadata_defs={'plaintext': 'uint8'})
    )
    assert isinstance(ths, estraces.TraceHeaderSet)
    stm = 's_ths.samples.shape'
    s_ths = ths[:100]
    t1 = timeit.timeit(stm, globals={'s_ths': s_ths}, number=10)
    s_ths = ths[:10]
    t2 = timeit.timeit(stm, globals={'s_ths': s_ths}, number=10)
    assert t1 / t2 < 10


# see https://gitlab.com/eshard/estraces/-/issues/45
def test_filename_property(filename):
    ths = estraces.read_ths_from_sqlite(
        filename,
        samples_config=sqlite_format.SamplesConfig(),
        metadata_config=sqlite_format.MetadataConfig(metadata_defs={'plaintext': 'uint8'})
    )
    assert ths.filename == test_path



# -*- coding: utf-8 -*-
from ..formats import bin_format as _bin_format
from ..traces.trace_header_set import build_trace_header_set
import warnings as _warnings


class _LegacyBinFormat(_bin_format.BinFormat):

    @property
    def filename(self):
        return self._filenames


def _bin_reader(filename_pattern, dtype, offset=0, **kwargs):
    _warnings.warn(
        'BinReader is deprecated. Use read_ths_from_bin functions from estraces package.',
        DeprecationWarning
    )

    files_list = _bin_format.get_sorted_filenames(pattern=filename_pattern)
    reader = _LegacyBinFormat(
        filenames=files_list,
        offset=offset,
        dtype=dtype,
        metadatas_parsers=kwargs,
    )
    return build_trace_header_set(
        reader=reader,
        name=f'Trace header set with {reader}.'
    )


BinReader = _bin_reader



from .context import estraces # noqa
import pytest
import shutil
import os
import numpy as np
import timeit
import memory_profiler
import h5py
from estraces.formats import ets_format
from pathlib import Path
import cloudpickle

path = 'tests/samples/test.ets'


@pytest.fixture
def ets1():
    return ets_format.read_ths_from_ets_file(filename=path)


def test_open_ets_file(ets1):
    assert isinstance(ets1, estraces.TraceHeaderSet)
    assert 200 == len(ets1)


def test_open_ets_file_with_path():
    ets1 = ets_format.read_ths_from_ets_file(filename=Path(path))
    assert isinstance(ets1, estraces.TraceHeaderSet)
    assert 200 == len(ets1)


def test_read_from_ets_raise_exception_if_file_does_not_exist():
    with pytest.raises(AttributeError):
        ets_format.read_ths_from_ets_file(filename='doesnotexist')


def test_read_from_incorrect_ets_file_raises_type_error():
    with pytest.raises(TypeError):
        ets_format.read_ths_from_ets_file(filename='tests/samples/AESEncrypt.500MSs.cur.00000')


def test_read_metadatas(ets1):
    assert isinstance(ets1.metadatas, estraces.traces.metadatas.Metadatas)
    for v in ets1.metadatas.keys():
        assert v in ['ciphertext', 'foo_bar', 'plaintext']
    assert [2, 9, 15, 8, 11, 8, 2, 13] == ets1.metadatas['plaintext'][0].tolist()
    assert [2, 9, 15, 8, 11, 8, 2, 13] == ets1.plaintext[0].tolist()
    assert [2, 9, 15, 8, 11, 8, 2, 13] == ets1[0].plaintext.tolist()
    for _, v in ets1.metadatas.items():
        assert 2 == v.ndim
    for _, v in ets1[0].metadatas.items():
        assert 1 == v.ndim
    assert [
        [3, 10, 16, 9, 12, 9, 3, 14],
        [4, 11, 17, 10, 13, 10, 4, 15],
        [5, 12, 18, 11, 14, 11, 5, 16]
    ] == ets1.plaintext[1: 4].tolist()


def test_read_samples(ets1):
    assert isinstance(ets1.samples, estraces.traces.samples.Samples)
    assert ets1.samples.ndim == 2
    assert isinstance(ets1[0].samples, estraces.traces.samples.Samples)
    assert ets1[0].samples.ndim == 1
    assert [
        [2, 11, 21],
        [4, 13, 23],
        [6, 15, 25]
    ] == ets1.samples[1: 6: 2, [1, 10, 20]].tolist()

    assert [
        [2, 11, 21],
        [4, 13, 23],
        [6, 15, 25]
    ] == ets1.samples[1: 6: 2, np.array([1, 10, 20])].tolist()

    assert [
        [2, 3, 4, 5, 6, 7, 8, 9, 10],
        [4, 5, 6, 7, 8, 9, 10, 11, 12],
        [6, 7, 8, 9, 10, 11, 12, 13, 14]
    ] == ets1.samples[1: 6: 2, 1:10].tolist()

    # Manage limits
    assert 0 == ets1.samples[0, 0]
    assert [1, 2, 3, 4, 5, 6, 7, 8, 9] == ets1.samples[0, 1:10].tolist()
    assert [1, 2, 3, 4, 5, 6, 7, 8, 9] == ets1.samples[1:10, 0].tolist()

    assert [1, 3, 4, 7, 9] == ets1[0].samples[np.array([1, 3, 4, 7, 9])].tolist()

    # H5 doesn't manage unordered or dedup list, arrays in constrast to ndarray.
    assert [1, 4, 3, 7, 4, 9] == ets1[0].samples[np.array([1, 4, 3, 7, 4, 9])].tolist()

    assert list(range(33)) == ets1[0].samples[...].tolist()
    assert 12 == ets1[0].samples[12]

    # Test memory overhead optimization strategies
    assert ets1.samples[:, :][:, [0, 32, 28, 0]].tolist() == ets1.samples[:, [0, 32, 28, 0]].tolist()


def test_read_samples_with_range(ets1):
    assert [
        [2, 3, 4, 5, 6, 7, 8, 9, 10],
        [4, 5, 6, 7, 8, 9, 10, 11, 12],
        [6, 7, 8, 9, 10, 11, 12, 13, 14]
    ] == ets1.samples[range(1, 6, 2), range(1, 10)].tolist()


def test_slice_ets(ets1):
    sub1 = ets1[2: 7]
    sub2 = ets1[[1, 10, 12]]
    assert isinstance(sub1, estraces.TraceHeaderSet)
    assert isinstance(sub2, estraces.TraceHeaderSet)
    assert 5 == len(sub1)
    assert 3 == len(sub2)
    assert [2, 3, 4, 5] == sub1.samples[0, 0:4].tolist()
    assert [6, 7, 8, 9] == sub1.samples[4, 0:4].tolist()
    assert [1, 2, 3, 4] == sub2.samples[0, 0:4].tolist()
    assert [10, 11, 12, 13] == sub2.samples[1, 0:4].tolist()
    assert [12, 13, 14, 15] == sub2.samples[2, 0:4].tolist()


def bench_func(ets):
    for t in ets:
        t.plaintext


def ref_func(ets):
    plaintext = ets.plaintext[:]
    for t in range(len(ets)):
        plaintext[t]


ets2 = ets_format.read_ths_from_ets_file(filename=path)
ets3 = ets_format.read_ths_from_ets_file(filename=path)


def test_fetch_samples_execution_time(ets1):
    # See issue https://gitlab.eshard.int/esdynamic/estraces/issues/26
    # Here we mesure execution time of bench_func versus an arbitrary ref time given by ref_func execution.
    # Initial ratio before issue #26 ~175
    t_ref = timeit.timeit('ref_func(ets2)', number=100, globals=globals())
    t = timeit.timeit('bench_func(ets3)', number=100, globals=globals())
    assert 200 > abs(t - t_ref) / t_ref


PATH = 'tests/samples/'


@pytest.fixture
def ets_filename():
    fn = 'big_ets'
    ets_filename = f'{PATH}{fn}.ets'
    ets_filename_bck = f'{PATH}{fn}.ets.bck'
    shutil.copyfile(ets_filename, ets_filename_bck)
    num_traces = 50000
    len_trc = 1000
    plains = np.array([np.random.randint(0, 256, 32, dtype="uint8") for i in range(num_traces)])
    datas = np.vstack([np.random.rand(len_trc) for i in range(num_traces)])

    file = h5py.File(name=ets_filename, mode='w')
    file.create_group('metadata')
    file.create_dataset('traces', dtype='float64', shape=(num_traces, len_trc))
    file['metadata'].create_dataset('plaintext', dtype='uint8', shape=(num_traces, 32))
    for i in range(num_traces):
        file['traces'][i] = datas[i]
        file['metadata']['plaintext'][i] = plains[i]
    file.flush()
    file.close()
    yield ets_filename
    shutil.copyfile(ets_filename_bck, ets_filename)


def test_ths_slicing_memory_consumption(ets_filename):
    # See issue https://gitlab.eshard.int/esdynamic/estraces/issues/30
    # Slicing a ths based on ets file without direct acces to data lead to important memory consumptions.
    # For example ths[0] or ths[0:100] can lead to the total size of the ets consummed in memory.
    # Here we test that slicing a ths with an integer or slice without access to the samples
    # doesn't load too much data === at most 10% memory consumption of the ths.samples reading operation.
    # We also ensure that slicing or getting only one trace has the same memory footprint.

    def tested_trace():
        ths = estraces.read_ths_from_ets_file(ets_filename)
        ths[0]

    trace_footprint = max(memory_profiler.memory_usage(tested_trace))

    def tested_slice():
        ths = estraces.read_ths_from_ets_file(ets_filename)
        ths[:10]

    slice_footprint = max(memory_profiler.memory_usage(tested_slice))

    def bgd():
        estraces.read_ths_from_ets_file(ets_filename)

    bgd_footprint = max(memory_profiler.memory_usage(bgd))

    def ref():
        ths = estraces.read_ths_from_ets_file(ets_filename)
        ths[:].samples[:]

    ref_footprint = max(memory_profiler.memory_usage(ref))
    assert abs(trace_footprint - bgd_footprint) / abs(ref_footprint - bgd_footprint) < 0.1, 'Memory consumption of trace selection is too high'
    assert abs(slice_footprint - bgd_footprint) / abs(ref_footprint - bgd_footprint) < 0.1, 'Memory consumption of ths slicing is too high'
    assert abs((trace_footprint - slice_footprint) / (slice_footprint)) < 0.01, 'Memory consumption of ths slicing and trace selection should be equivalent'


def test_split_preserves_length_consistency_of_sub_ths(ets1):
    parts = ets1.split(100)
    for _, p in enumerate(parts):
        assert len(p) in (100, 56)
        assert len(p) == p.samples[:].shape[0]


@pytest.fixture
def ets_with_name():
    data = np.random.randint(0, 255, (1000, 2000), dtype='uint8')
    name = np.array(['a great name' for name in range(1000)])
    _id = np.array([i for i in range(1000)], dtype='uint')
    ths = estraces.read_ths_from_ram(data, name=name, id=_id)
    out = estraces.ETSWriter('ets_with_name.ets')
    out.add_trace_header_set(ths)
    try:
        yield out.get_reader()
    finally:
        os.remove('ets_with_name.ets')


def test_ets_with_name_and_id_metadata(ets_with_name):
    assert ets_with_name.name[0] == 'a great name'
    assert len(ets_with_name.name) == 1000
    assert ets_with_name[0].name == 'a great name'
    assert len(ets_with_name.id) == 1000
    assert ets_with_name.id[0] == 0
    assert ets_with_name[100].id == 100


def test_headers():
    ets = ets_format.read_ths_from_ets_file('tests/samples/test_with_headers.ets')
    headers_dict = {'foo': 'bar', 'bar': 42, 'list': [1, 2, 3], 'ndarray': np.arange(5, dtype='int8')}
    assert ets.headers == headers_dict


def test_serialization(ets1):
    ths = ets1
    dumped = cloudpickle.dumps(ths)
    loaded_ths = cloudpickle.loads(dumped)
    assert np.array_equal(loaded_ths.samples[:], ths.samples[:])
    assert np.array_equal(loaded_ths.plaintext[:], ths.plaintext[:])
    assert np.array_equal(loaded_ths.foo_bar[:], ths.foo_bar[:])
    assert np.array_equal(loaded_ths.ciphertext[:], ths.ciphertext[:])


@pytest.mark.parametrize('ets_file', ['tests/samples/test_old.ets', 'tests/samples/test_new.ets'])
def test_read_with_strings(ets_file):
    ets = ets_format.read_ths_from_ets_file(ets_file)
    for metadata_name in ['string', 'variable_length_string', 'ascii_bytes', 'variable_length_ascii_bytes']:
        print(metadata_name, end=' ')
        metadata = getattr(ets, metadata_name)
        for value in metadata:
            assert isinstance(value, str)
        print('is of type ', type(value))


# see https://gitlab.com/eshard/estraces/-/issues/45
def test_filename_property(ets1):
    assert ets1.filename == path



# -*- coding: utf-8 -*-
from . import metadatas
from . import samples
from . import headers


class Trace:
    """Provide a consistent API to manipulate a trace samples data and metadatas.

    Attributes:
        samples (:class:`Samples`): 1 dimension samples data
        metadatas (:class:`Metadatas`): trace metadatas
        headers (:class:`Headers`): headers value of the trace set
    Note:
        All metadatas are available through the `metadatas` attributes and through a corresponding named property.

    Examples:
        Samples are available as 1 dimensional array-like :class:`Samples` with shape (size of trace,).
        It supports a subset of numpy.ndarray-like slicing - including advanced list-based slicing::

            trace.samples[2000:3000]
            trace.samples[[1, 100, 1000]] # get samples at 1, 100 and 1000 indexes

        Metadatas are available through a dict-like :class:`Metadatas` instance::

            metadatas = trace.metadatas
            metadatas['plaintext']

        Each metadata can be reached with its own property::

            trace.plaintext # is equivalent to trace.metadatas['plaintext']

        Headers are metadata value which are the same for all the traces of one given trace set.
        It is provided at the trace level through a dict-like object:

            trace.headers['key'] # equivalent to ths.headers['key'] where ths is the trace header set of the trace.

    """

    _is_initialized = False

    def __init__(self, trace_id, reader):
        if trace_id is None:
            raise AttributeError("trace_id can't be None.")
        self._id = trace_id
        self._reader = reader
        self._samples = None
        self._metadatas = None
        self._headers = None
        try:
            self.name = f'Trace n°{self._id}'
        except AttributeError:
            # Name is already a metadata of the ths, so we don't bother with it.
            pass
        try:
            self.id = self._id
        except AttributeError:
            # id is already a metadata of the ths, so we don't bother with it.
            pass
        self._is_initialized = True

    def __len__(self):
        return len(self.samples)

    @property
    def samples(self):
        if self._samples is None:
            self._samples = samples.Samples(reader=self._reader, trace_id=self._id)
        return self._samples

    @property
    def metadatas(self):
        if not self._metadatas:
            self._metadatas = metadatas.Metadatas(reader=self._reader, trace_id=self._id)
        return self._metadatas

    @property
    def headers(self):
        if not self._headers:
            self._headers = headers.Headers(reader=self._reader)
        return self._headers

    def __repr__(self):
        return str(self)

    def __str__(self):
        r = 'Trace:\n'
        r += f'{"Reader instance":.<17}: {self._reader}\n'
        r += f'{"Index in set":.<17}: {self._id}\n'
        r += f'{"Samples size":.<17}: {len(self)}\n'
        for k in self.metadatas.keys():
            r += f'{k:.<17}: {self.metadatas.get(k)}\n'
        return r

    def __getattr__(self, name):
        try:
            return self.metadatas[name]
        except KeyError:
            raise AttributeError(f'No attribute {name}.')

    def __setattr__(self, name, value):
        if name not in self.__dir__() and self._is_initialized:
            self.metadatas  # Metadata needs to be instantiated before adding a in-memory metadata.
            self._metadatas[name] = value
        super().__setattr__(name, value)



import psutil as _ps
import os as _os
import h5py as _h5py
import numpy as _np
from .ets_format import read_ths_from_ets_file
from ..traces import samples as _samples, metadatas as _metadata, trace as _trace, trace_header_set as _ths
import logging
import warnings
from pathlib import Path as _Path


logger = logging.getLogger(__name__)

METADATA_GROUP_KEY = 'metadata'
TRACES_DATASET_KEY = 'traces'


class ETSWriterError(Exception):
    pass


class ETSWriter:
    """Provide API to create an ETS (Eshard Trace Set) file."""

    def __init__(self, filename, overwrite=False):
        """Create an ETS file writer instance.

        Args:
            filename (str): path and filename to write the ETS.
            overwrite (bool, default=False): if True, any existing file with filename will be erased before writing data.

        """
        self._overwrite = overwrite
        self._filename = filename
        self._h5_file = None
        self._initialized_datasets = None
        if _os.path.isfile(self._filename):
            if self._overwrite:
                logger.warning(f'File {self._filename} already exists. It will be reset on the first write operation.')
            else:
                logger.info(f'File {self._filename} already exists. If no indexes are specified, new data will be append on existing datasets.')
        self._is_init = False

    def _close_all(self):
        try:
            if not _os.path.exists(self._filename):
                return
            f = _h5py.File(self._filename, 'r')
            to_close = _h5py.h5f.get_obj_ids(f.id)
            for fid in to_close:
                tmp = _h5py.File(fid, 'r')
                tmp.flush()
                tmp.close()
            f.close()
        except OSError:
            pass

    def _init_file(self):
        self._close_all()
        if _os.path.isfile(self._filename) and self._overwrite:
            _os.remove(self._filename)
            self._h5_file = _h5py.File(self._filename, mode='w', libver='latest')
        else:
            self._h5_file = _h5py.File(self._filename, mode='a', libver='latest')

        self._h5_file.swmr_mode = True

        if METADATA_GROUP_KEY not in list(self._h5_file.keys()):
            self._h5_file.create_group(METADATA_GROUP_KEY)
        self._initialized_datasets = list(self._h5_file[METADATA_GROUP_KEY].keys())
        self._is_init_traces = TRACES_DATASET_KEY in list(self._h5_file.keys())
        self._is_init = True

    def __del__(self):
        self.close()

    @staticmethod
    def _np_to_str(data):
        s = data.shape
        data = str(data)
        if s != ():
            data = data[1:-1]
        return _np.array([data])

    def add_trace_header_set(self, trace_header_set: _ths.TraceHeaderSet):
        """Append all traces samples and metadata hold by trace_header_set at the end of the ETS.

        Args:
            trace_header_set (`TraceHeaderSet`): a trace header set instance.

        """
        if not isinstance(trace_header_set, _ths.TraceHeaderSet):
            raise TypeError(f'trace header set is not a TraceHeaderSet instance but {type(trace_header_set)}.')
        if self._initialized_datasets and self._initialized_datasets != list(trace_header_set.metadatas.keys()):
            raise ETSWriterError(
                f'trace header set instance has different metadata {trace_header_set.metadatas.keys()} than existing in file {self._initialized_datasets}.'
            )

        batch_size = _get_optimal_batch_size(trace_header_set)
        self.add_samples(trace_header_set.samples, batch_size=batch_size)
        self.add_metadata(trace_header_set.metadatas, batch_size=batch_size)
        self.write_headers(trace_header_set.headers)

    def add_trace(self, trace: _trace.Trace):
        """Append trace samples and metadata provided to the end of the ETS.

        Args:
            trace (`Trace`): a Trace instance.

        """
        if not isinstance(trace, _trace.Trace):
            raise TypeError(f'trace is not a Trace instance but {type(trace)}.')
        if self._initialized_datasets and self._initialized_datasets != list(trace.metadatas.keys()):
            raise ETSWriterError(f'trace instance has different metadata {trace.metadatas.keys()} than existing in file {self._initialized_datasets}.')
        self.add_samples(trace.samples)
        self.add_metadata(trace.metadatas)

    def add_samples(self, samples: _samples.Samples, batch_size=1):
        """Append provided samples to the end of ETS.

        Warning: when using directly this method, it is up to you to ensure consistency of indices between samples and metadata in
        your ETS file. Use of `add_trace` or `add_trace_header_set` methods is recommended.

        Args:
            samples (`Samples`): a samples instance.

        """
        if not isinstance(samples, _samples.Samples):
            raise TypeError(f'samples is not a Samples instance but {type(samples)}.')
        if samples.ndim > 1:
            for i in range(0, len(samples), batch_size):
                self.write_samples(samples[i: i + batch_size])
        else:
            self.write_samples(samples[:])

    def add_metadata(self, metadata: _metadata.Metadatas, batch_size=1):
        """Append provided metadata to the end of ETS.

        Warning: when using directly this method, it is up to you to ensure consistency of indices between samples and metadata in
        your ETS file. Use of `add_trace` or `add_trace_header_set` methods is recommended.

        Args:
            metadata (`Metadatas`): a metadata instance.

        """
        if not isinstance(metadata, _metadata.Metadatas):
            raise TypeError(f'metadata is not a Metadatas instance, not {type(metadata)}.')
        for key, val in metadata.items():
            if metadata.is_trace():
                self.write_metadata(key, val)
            else:
                for i in range(0, len(val), batch_size):
                    self.write_metadata(key, val[i: i + batch_size], scalar=val.ndim != 2)

    def write_trace_object_and_points(self, trace_object, points, index=None):
        """Write provided trace samples and metadata at specified index of the ETS.

        If index is None, data will be appended to the end of the ETS.

        Args:
            trace_object (`Trace`): a Trace instance.
            points (numpy.ndarray): samples to write.
            index (int, default=None): index at which to write.

        """
        for tag, value in trace_object.metadatas.items():
            self.write_metadata(tag, value, index=index)
        self.write_samples(points, index=index)

    def write_samples(self, samples_array, index=None):
        """Write provided trace samples at specified index of the ETS.

        If index is None, data will be appended to the end of the ETS.

        Warning: when using directly this method, it is up to you to ensure consistency of indices between samples and metadata in
        your ETS file.

        Args:
            samples_array (numpy.ndarray): samples to write.
            index (int, default=None): index at which to write.

        """
        if not self._is_init:
            self._init_file()
        if not self._is_init_traces:
            self._init_traces(samples_array)
        self._add_to_dataset(self._h5_file[TRACES_DATASET_KEY], samples_array, index)

    def write_points(self, points, index=None):
        warnings.warn('This method is deprecated and will be removed in a future version. Use write_samples instead.', DeprecationWarning)
        return self.write_samples(points, index=index)

    def write_headers(self, headers):
        if not self._is_init:
            self._init_file()
        self._h5_file[METADATA_GROUP_KEY].attrs.update(headers)

    def write_metadata(self, key, value, index=None, scalar=False):
        """Write provided metadata at specified index of the ETS.

        If index is None, data will be appended to the end of the ETS.

        Warning: when using directly this method, it is up to you to ensure consistency of indices between samples and metadata in
        your ETS file.

        Args:
            key (str): metadata key
            value (numpy.ndarray): metadata value to write.
            index (int, default=None): index at which to write.

        """
        if not self._is_init:
            self._init_file()
        if not isinstance(value, _np.ndarray):
            value = _np.array(value)
        if key not in self._initialized_datasets:
            self._init_metadata(key, value, scalar=scalar)
        self._add_to_dataset(self._h5_file[METADATA_GROUP_KEY][key], value, index, scalar=scalar)

    def write_meta(self, tag, metadata, index=None):
        warnings.warn('This method is deprecated and will be removed in a future version. Use write_samples instead.', DeprecationWarning)
        return self.write_metadata(tag, metadata, index=index)

    @staticmethod
    def _compute_shapes_and_htype(array):
        if array.ndim == 1:
            shape = (0, len(array))
        elif array.ndim == 2:
            shape = (0, array.shape[1])
        else:
            shape = (0, 1)
        return shape, (None, shape[1]), array.dtype

    def _init_traces(self, data):
        shape, maxshape, dtype = self._compute_shapes_and_htype(data)
        self._h5_file.create_dataset(TRACES_DATASET_KEY, shape, maxshape=maxshape, dtype=dtype)
        self._is_init_traces = True

    def _init_metadata(self, metadata_key, metadata, scalar=False):
        shape, maxshape, h5type = self._compute_shapes_and_htype(metadata)

        if metadata.dtype.kind not in ['b', 'i', 'u', 'f', 'c']:
            h5type = _h5py.special_dtype(vlen=str)
            shape = (0, )
            maxshape = (None,)
        elif scalar:
            shape = (0, )
            maxshape = (None, )

        self._h5_file[METADATA_GROUP_KEY].create_dataset(metadata_key, shape, maxshape=maxshape, dtype=h5type)
        self._initialized_datasets.append(metadata_key)

    def _add_to_dataset(self, data_set, data, index, scalar=False):
        if data.ndim == 1 and not scalar:
            _data = _np.empty((1, len(data)), dtype=data.dtype)
            _data[0, :] = data
        elif (scalar and data.ndim == 1) or data.ndim == 2:
            _data = data
        else:
            _data = _np.array([[data]])
        return self._write_to_data_set(data_set, _data, index)

    def _write_to_data_set(self, data_set, data, index):
        if index is None:
            index = data_set.shape[0]
        if index < data_set.shape[0] and not self._overwrite:
            raise ETSWriterError(
                f'An element already exists in {str(data_set.name)} dataset at index {index} and overwriting is disabled.')
        if index >= data_set.shape[0]:
            data_set.resize(index + data.shape[0], axis=0)

        if data_set.ndim == 2 and data.ndim == 2 and data.shape[1] != data_set.shape[1]:
            shape = (data.shape[0], data_set.shape[1])
            tmp_data = _np.zeros(shape, dtype=data.dtype)
            data = data[:, :data_set.shape[1]]
            tmp_data[:, :data.shape[1]] = data
            tmp_data[:, data.shape[1]:] = 0
            data = tmp_data

        data_set[index: index + data.shape[0]] = data
        self._h5_file.flush()

    def get_reader(self):
        """Returns a `TraceHeaderSet` instance from the current writer, and closes it."""
        self.close()
        return read_ths_from_ets_file(self._filename)

    def close(self):
        """Close the current instance ETS file."""
        if self._h5_file:
            self._h5_file.close()


def compress_ets(filename, out_filename):
    """Create a compressed version of an ETS file.

    Creates a new ETS file using compression options provided by h5.
    The gain of compression can highly vary, depending on the nature of the data, from 10% to 50%.
    The trade-off is a somewhat slower reading from the resulting file.

    Args:
        filename (str, pathlib.Path): filename of original uncompressed ETS file.
        out_filename (str, pathlib.Path): filename of output compressed ETS file.

    Raises:
        ValueError if a file named out_filename already exists.
        AttributeError if the original file filename doesn't exists.

    """
    if not isinstance(filename, (str, _Path)) or not isinstance(out_filename, (str, _Path)):
        raise TypeError(f'filename and out_filename must be str or Path, not {type(filename)}, {type(out_filename)}.')

    if _os.path.exists(out_filename):
        raise ValueError(f'A file named {out_filename} already exists.')
    ths = read_ths_from_ets_file(filename)

    ets_compressed = _h5py.File(out_filename, mode='w', libver='latest')
    ets_compressed.create_dataset(
        TRACES_DATASET_KEY,
        shape=(len(ths), len(ths.samples[0])),
        dtype=ths.samples[0].dtype,
        compression='gzip',
        compression_opts=9
    )
    ets_compressed.create_group(METADATA_GROUP_KEY)
    ets_compressed[METADATA_GROUP_KEY].attrs.update(dict(ths.headers))
    batch_size = _get_optimal_batch_size(ths)

    for i in range(0, len(ths), batch_size):
        ets_compressed[TRACES_DATASET_KEY][i: i + batch_size] = ths[i: i + batch_size].samples[:]
        for k, v in ths[i: i + batch_size].metadatas.items():
            if v.dtype.kind not in ['b', 'i', 'u', 'f', 'c']:
                h5type = _h5py.special_dtype(vlen=str)
                shape = (len(ths), )
            else:
                h5type = v.dtype
                shape = (len(ths), v.shape[-1])

            if k not in ets_compressed[METADATA_GROUP_KEY]:
                ets_compressed[METADATA_GROUP_KEY].create_dataset(k, shape=shape, compression='gzip', compression_opts=9, dtype=h5type)
            ets_compressed[METADATA_GROUP_KEY][k][i: i + batch_size] = v

    ets_compressed.close()


def _get_optimal_batch_size(ths, available_memory_ratio=0.5, limit=8e9):
    base_trace = ths[0]
    unit_size = base_trace.samples[:].nbytes
    for k, v in base_trace.metadatas.items():
        try:
            unit_size += v.nbytes
        except AttributeError:
            logger.info(f'Scalar metadata {k} not taken into account for optimal batch size computation.')
    available_mem = _ps.virtual_memory()[1]

    if unit_size > available_mem:
        raise MemoryError(f"Not enough memory to load one trace (Available memory : {available_mem} bytes)")
    return int(min(available_memory_ratio * available_mem, limit) / unit_size)
